# Copyright (C) 2002-2013 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2003-2011 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2013 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


import atexit, errno, functools, glob, inspect, io, itertools, logging, os
import queue, re, shutil, sys, tempfile, textwrap, threading

from pyx import config, unit, box, baseclasses, trafo, version, attr, style, path, canvas
from pyx import bbox as bboxmodule
from pyx.dvi import dvifile

logger = logging.getLogger("pyx")


def indent_text(text):
    "Prepends two spaces to each line in text."
    return "".join("  " + line for line in text.splitlines(True))


def remove_string(p, s):
    """Removes a string from a string.

    The function removes the first occurrence of a string in another string.

    :param str p: string to be removed
    :param str s: string to be searched
    :returns: tuple of the result string and a success boolean (``True`` when
        the string was removed)
    :rtype: tuple of str and bool

    .. testsetup::

        import re
        from pyx.text import *

    Example:
        >>> remove_string("XXX", "abcXXXdefXXXghi")
        ('abcdefXXXghi', True)

    """
    r = s.replace(p, '', 1)
    return r, r != s


def remove_pattern(p, s, ignore_nl=True):
    r"""Removes a pattern from a string.

    The function removes the first occurence of the pattern from a string. It
    returns a tuple of the resulting string and the matching object (or
    ``None``, if the pattern did not match).

    :param re.regex p: pattern to be removed
    :param str s: string to be searched
    :param bool ignore_nl: When ``True``, newlines in the string are ignored
        during the pattern search. The returned string will still contain all
        newline characters outside of the matching part of the string, whereas
        the returned matching object will not contain the newline characters
        inside of the matching part of the string.
    :returns: the result string and the match object or ``None`` if
        search failed
    :rtype: tuple of str and (re.match or None)

    Example:
        >>> r, m = remove_pattern(re.compile("XXX"), 'ab\ncXX\nXdefXX\nX')
        >>> r
        'ab\ncdefXX\nX'
        >>> m.string[m.start():m.end()]
        'XXX'

    """
    if ignore_nl:
        r = s.replace('\n', '')
        has_nl = r != s
    else:
        r = s
        has_nl = False
    m = p.search(r)
    if m:
        s_start = r_start = m.start()
        s_end = r_end = m.end()
        if has_nl:
            j = 0
            for c in s:
                if c == '\n':
                    if j < r_end:
                        s_end += 1
                        if j <= r_start:
                            s_start += 1
                else:
                    j += 1
        return s[:s_start] + s[s_end:], m
    return s, None


def index_all(c, s):
    """Return list of positions of a character in a string.

    Example:
        >>> index_all("X", "abXcdXef")
        [2, 5]

    """
    assert len(c) == 1
    return [i for i, x in enumerate(s) if x == c]


def pairwise(i):
    a, b = itertools.tee(i)
    next(b, None)
    return zip(a, b)


def remove_nested_brackets(s, openbracket="(", closebracket=")", quote='"'):
    """Remove nested brackets

    Return a modified string with all nested brackets 1 removed, i.e. only
    keep the first bracket nesting level. In case an opening bracket is
    immediately followed by a quote, the quoted string is left untouched,
    even if it contains brackets. The use-case for that are files in the
    folder "Program Files (x86)".

    If the bracket nesting level is broken (unbalanced), the unmodified
    string is returned.

    Example:
        >>> remove_nested_brackets('aaa("bb()bb" cc(dd(ee))ff)ggg'*2)
        'aaa("bb()bb" ccff)gggaaa("bb()bb" ccff)ggg'

    """
    openpos = index_all(openbracket, s)
    closepos = index_all(closebracket, s)
    if quote is not None:
        quotepos = index_all(quote, s)
        for openquote, closequote in pairwise(quotepos):
             if openquote-1 in openpos:
                 # ignore brackets in quoted string
                 openpos = [pos for pos in openpos
                            if not (openquote < pos < closequote)]
                 closepos = [pos for pos in closepos
                             if not (openquote < pos < closequote)]
    if len(openpos) != len(closepos):
        # unbalanced brackets
        return s

    # keep the original string in case we need to return due to broken nesting levels
    r = s

    level = 0
    # Iterate over the bracket positions from the end.
    # We go reversely to be able to immediately remove nested bracket levels
    # without influencing bracket positions yet to come in the loop.
    for pos, leveldelta in sorted(itertools.chain(zip(openpos, itertools.repeat(-1)),
                                                  zip(closepos, itertools.repeat(1))),
                                  reverse=True):
        # the current bracket nesting level
        level += leveldelta
        if level < 0:
            # unbalanced brackets
            return s
        if leveldelta == 1 and level == 2:
            # a closing bracket to cut after
            endpos = pos+1
        if leveldelta == -1 and level == 1:
            # an opening bracket to cut at -> remove
            r = r[:pos] + r[endpos:]
    return r


class TexResultError(ValueError):
    "Error raised by :class:`texmessage` parsers."
    pass


class texmessage:
    """Collection of TeX output parsers.

    This class is not meant to be instanciated. Instead, it serves as a
    namespace for TeX output parsers, which are functions receiving a TeX
    output and returning parsed output.

    In addition, this class also contains some generator functions (namely
    :attr:`texmessage.no_file` and :attr:`texmessage.pattern`), which return a
    function according to the given parameters. They are used to generate some
    of the parsers in this class and can be used to create others as well.
    """

    start_pattern = re.compile(r"This is [-0-9a-zA-Z\s_]*TeX")

    @staticmethod
    def start(msg):
        r"""Validate TeX/LaTeX startup message including scrollmode test.

        Example:
            >>> texmessage.start(r'''
            ... This is e-TeX (version)
            ... *! Undefined control sequence.
            ... <*> \raiseerror
            ...                %
            ... ''')
            ''

        """
        # check for "This is e-TeX" etc.
        if not texmessage.start_pattern.search(msg):
            raise TexResultError("TeX startup failed")

        # check for \raiseerror -- just to be sure that communication works
        new = msg.split("*! Undefined control sequence.\n<*> \\raiseerror\n               %\n", 1)[-1]
        if msg == new:
            raise TexResultError("TeX scrollmode check failed")
        return new

    @staticmethod
    def no_file(fileending, qualname=None):
        "Generator function to ignore the missing file message for fileending."
        def check(msg):
            "Ignore the missing {} file message."
            return msg.replace("No file texput.%s." % fileending, "").replace("No file %s%stexput.%s." % (os.curdir, os.sep, fileending), "")
        check.__doc__ = check.__doc__.format(fileending)
        if qualname is not None:
            check.__qualname__ = qualname
        return check

    no_aux = staticmethod(no_file.__func__("aux", "texmessage.no_aux"))
    no_nav = staticmethod(no_file.__func__("nav", "texmessage.no_nav"))

    aux_pattern = re.compile(r'\(([^()]+\.aux|"[^"]+\.aux")\)')
    log_pattern = re.compile(r'Transcript written on .*texput\.log"?\.', re.DOTALL)

    @staticmethod
    def end(msg):
        "Validate TeX shutdown message."
        msg = re.sub(texmessage.aux_pattern, "", msg).replace("(see the transcript file for additional information)", "")

        # check for "Transcript written on ...log."
        msg, m = remove_pattern(texmessage.log_pattern, msg)
        if not m:
            raise TexResultError("TeX logfile message expected")
        return msg

    quoted_file_pattern = re.compile(r'\("(?P<filename>[^"]+)".*?\)')
    file_pattern = re.compile(r'\((?P<filename>[^"][^ )]*).*?\)', re.DOTALL)

    @staticmethod
    def load(msg):
        """Ignore file loading messages.

        Removes text starting with a round bracket followed by a filename
        ignoring all further text until the corresponding closing bracket.
        Quotes and/or line breaks in the filename are handled as needed for TeX
        output.

        Without quoting the filename, the necessary removal of line breaks is
        not well defined and the different possibilities are tested to check
        whether one solution is ok. The last of the examples below checks this
        behavior.

        Examples:
            >>> texmessage.load(r'''other (text.py) things''')
            'other  things'
            >>> texmessage.load(r'''other ("text.py") things''')
            'other  things'
            >>> texmessage.load(r'''other ("tex
            ... t.py" further (ignored)
            ... text) things''')
            'other  things'
            >>> texmessage.load(r'''other (t
            ... ext
            ... .py
            ... fur
            ... ther (ignored) text) things''')
            'other  things'

        """
        r = remove_nested_brackets(msg)
        r, m = remove_pattern(texmessage.quoted_file_pattern, r)
        while m:
            if not os.path.isfile(config.get("text", "chroot", "") + m.group("filename")):
                return msg
            r, m = remove_pattern(texmessage.quoted_file_pattern, r)
        r, m = remove_pattern(texmessage.file_pattern, r, ignore_nl=False)
        while m:
            for filename in itertools.accumulate(m.group("filename").split("\n")):
                if os.path.isfile(config.get("text", "chroot", "") + filename):
                    break
            else:
                return msg
            r, m = remove_pattern(texmessage.file_pattern, r, ignore_nl=False)
        return r

    quoted_def_pattern = re.compile(r'\("(?P<filename>[^"]+\.(fd|def))"\)')
    def_pattern = re.compile(r'\((?P<filename>[^"][^ )]*\.(fd|def))\)')

    @staticmethod
    def load_def(msg):
        "Ignore font definition (``*.fd`` and ``*.def``) loading messages."
        r = msg
        for p in [texmessage.quoted_def_pattern, texmessage.def_pattern]:
            r, m = remove_pattern(p, r)
            while m:
                if not os.path.isfile(config.get("text", "chroot", "") + m.group("filename")):
                    return msg
                r, m = remove_pattern(p, r)
        return r

    quoted_graphics_pattern = re.compile(r'<"(?P<filename>[^"]+\.eps)">')
    graphics_pattern = re.compile(r'<(?P<filename>[^"][^>]*\.eps)>')

    @staticmethod
    def load_graphics(msg):
        "Ignore graphics file (``*.eps``) loading messages."
        r = msg
        for p in [texmessage.quoted_graphics_pattern, texmessage.graphics_pattern]:
            r, m = remove_pattern(p, r)
            while m:
                if not os.path.isfile(config.get("text", "chroot", "") + m.group("filename")):
                    return msg
                r, m = remove_pattern(texmessage.quoted_file_pattern, r)
        return r

    @staticmethod
    def ignore(msg):
        """Ignore all messages.

        Should be used as a last resort only. You should write a proper TeX
        output parser function for the output you observe.

        """
        return ""

    @staticmethod
    def warn(msg):
        """Warn about all messages.

        Similar to :attr:`ignore`, but writing a warning to the logger about
        the TeX output. This is considered to be better when you need to get it
        working quickly as you will still be prompted about the unresolved
        output, while the processing continues.

        """
        if msg:
             logger.warning("ignoring TeX warnings:\n%s" % indent_text(msg.rstrip()))
        return ""

    @staticmethod
    def pattern(p, warning, qualname=None):
        "Warn by regular expression pattern matching."
        def check(msg):
            "Warn about {}."
            msg, m = remove_pattern(p, msg, ignore_nl=False)
            while m:
                logger.warning("ignoring %s:\n%s" % (warning, m.string[m.start(): m.end()].rstrip()))
                msg, m = remove_pattern(p, msg, ignore_nl=False)
            return msg
        check.__doc__ = check.__doc__.format(warning)
        if qualname is not None:
            check.__qualname__ = qualname
        return check

    box_warning = staticmethod(pattern.__func__(re.compile(r"^(Overfull|Underfull) \\[hv]box.*$(\n^..*$)*\n^$\n", re.MULTILINE),
                               "overfull/underfull box", qualname="texmessage.box_warning"))
    font_warning = staticmethod(pattern.__func__(re.compile(r"^LaTeX Font Warning: .*$(\n^\(Font\).*$)*", re.MULTILINE),
                                "font substitutions of NFSS", qualname="texmessage.font_warning"))
    package_warning = staticmethod(pattern.__func__(re.compile(r"^package\s+(?P<packagename>\S+)\s+warning\s*:[^\n]+(?:\n\(?(?P=packagename)\)?[^\n]*)*", re.MULTILINE | re.IGNORECASE),
                                   "generic package messages", qualname="texmessage.package_warning"))
    rerun_warning = staticmethod(pattern.__func__(re.compile(r"^(LaTeX Warning: Label\(s\) may have changed\. Rerun to get cross-references right\s*\.)$", re.MULTILINE),
                                 "rerun required message", qualname="texmessage.rerun_warning"))
    nobbl_warning = staticmethod(pattern.__func__(re.compile(r"^[\s\*]*(No file .*\.bbl.)\s*", re.MULTILINE),
                                 "no-bbl message", qualname="texmessage.nobbl_warning"))


###############################################################################
# textattrs
###############################################################################

_textattrspreamble = ""

class textattr:
    "a textattr defines a apply method, which modifies a (La)TeX expression"

class _localattr: pass

_textattrspreamble += r"""\gdef\PyXFlushHAlign{0}%
\def\PyXragged{%
\leftskip=0pt plus \PyXFlushHAlign fil%
\rightskip=0pt plus 1fil%
\advance\rightskip0pt plus -\PyXFlushHAlign fil%
\parfillskip=0pt%
\pretolerance=9999%
\tolerance=9999%
\parindent=0pt%
\hyphenpenalty=9999%
\exhyphenpenalty=9999}%
"""

class boxhalign(attr.exclusiveattr, textattr, _localattr):

    def __init__(self, aboxhalign):
        self.boxhalign = aboxhalign
        attr.exclusiveattr.__init__(self, boxhalign)

    def apply(self, expr):
        return r"\gdef\PyXBoxHAlign{%.5f}%s" % (self.boxhalign, expr)

boxhalign.left = boxhalign(0)
boxhalign.center = boxhalign(0.5)
boxhalign.right = boxhalign(1)
# boxhalign.clear = attr.clearclass(boxhalign) # we can't defined a clearclass for boxhalign since it can't clear a halign's boxhalign


class flushhalign(attr.exclusiveattr, textattr, _localattr):

    def __init__(self, aflushhalign):
        self.flushhalign = aflushhalign
        attr.exclusiveattr.__init__(self, flushhalign)

    def apply(self, expr):
        return r"\gdef\PyXFlushHAlign{%.5f}\PyXragged{}%s" % (self.flushhalign, expr)

flushhalign.left = flushhalign(0)
flushhalign.center = flushhalign(0.5)
flushhalign.right = flushhalign(1)
# flushhalign.clear = attr.clearclass(flushhalign) # we can't defined a clearclass for flushhalign since it couldn't clear a halign's flushhalign


class halign(boxhalign, flushhalign, _localattr):

    def __init__(self, aboxhalign, aflushhalign):
        self.boxhalign = aboxhalign
        self.flushhalign = aflushhalign
        attr.exclusiveattr.__init__(self, halign)

    def apply(self, expr):
        return r"\gdef\PyXBoxHAlign{%.5f}\gdef\PyXFlushHAlign{%.5f}\PyXragged{}%s" % (self.boxhalign, self.flushhalign, expr)

    def apply_trafo(self, textbox):
        return textbox.transform(trafo.translate_pt(-self.boxhalign*textbox.bbox().width_pt(), 0), keep_anchor=True)

halign.left = halign(0, 0)
halign.center = halign(0.5, 0.5)
halign.right = halign(1, 1)
halign.clear = attr.clearclass(halign)
halign.boxleft = boxhalign.left
halign.boxcenter = boxhalign.center
halign.boxright = boxhalign.right
halign.flushleft = halign.raggedright = flushhalign.left
halign.flushcenter = halign.raggedcenter = flushhalign.center
halign.flushright = halign.raggedleft = flushhalign.right


class _mathmode(attr.exclusiveattr, textattr, _localattr):
    "math mode"

    def __init__(self):
        attr.exclusiveattr.__init__(self, _mathmode)

    def apply(self, expr):
        return r"$\displaystyle{%s}$" % expr

    def apply_trafo(self, textbox):
        pass

mathmode = _mathmode()
clearmathmode = attr.clearclass(_mathmode)


class _phantom(attr.attr, textattr, _localattr):
    "phantom text"

    def apply(self, expr):
        return r"\phantom{%s}" % expr

phantom = _phantom()
clearphantom = attr.clearclass(_phantom)


_textattrspreamble += "\\newbox\\PyXBoxVBox%\n\\newdimen\\PyXDimenVBox%\n"

class parbox_pt(attr.sortbeforeexclusiveattr, textattr):

    top = 1
    middle = 2
    bottom = 3

    def __init__(self, width, baseline=top):
        self.width = width * 72.27 / (unit.scale["x"] * 72)
        self.baseline = baseline
        attr.sortbeforeexclusiveattr.__init__(self, parbox_pt, [_localattr])

    def apply(self, expr):
        if self.baseline == self.top:
            return r"\linewidth=%.5ftruept\vtop{\hsize=\linewidth\textwidth=\linewidth{}%s}" % (self.width, expr)
        elif self.baseline == self.middle:
            return r"\linewidth=%.5ftruept\setbox\PyXBoxVBox=\hbox{{\vtop{\hsize=\linewidth\textwidth=\linewidth{}%s}}}\PyXDimenVBox=0.5\dp\PyXBoxVBox\setbox\PyXBoxVBox=\hbox{{\vbox{\hsize=\linewidth\textwidth=\linewidth{}%s}}}\advance\PyXDimenVBox by -0.5\dp\PyXBoxVBox\lower\PyXDimenVBox\box\PyXBoxVBox" % (self.width, expr, expr)
        elif self.baseline == self.bottom:
            return r"\linewidth=%.5ftruept\vbox{\hsize=\linewidth\textwidth=\linewidth{}%s}" % (self.width, expr)
        else:
            ValueError("invalid baseline argument")

parbox_pt.clear = attr.clearclass(parbox_pt)

class parbox(parbox_pt):

    def __init__(self, width, **kwargs):
        parbox_pt.__init__(self, unit.topt(width), **kwargs)

parbox.clear = parbox_pt.clear


_textattrspreamble += "\\newbox\\PyXBoxVAlign%\n\\newdimen\\PyXDimenVAlign%\n"

class valign(attr.sortbeforeexclusiveattr, textattr):

    def __init__(self, avalign):
        self.valign = avalign
        attr.sortbeforeexclusiveattr.__init__(self, valign, [parbox_pt, _localattr])

    def apply(self, expr):
        return r"\setbox\PyXBoxVAlign=\hbox{{%s}}\PyXDimenVAlign=%.5f\ht\PyXBoxVAlign\advance\PyXDimenVAlign by -%.5f\dp\PyXBoxVAlign\lower\PyXDimenVAlign\box\PyXBoxVAlign" % (expr, 1-self.valign, self.valign)

valign.top = valign(0)
valign.middle = valign(0.5)
valign.bottom = valign(1)
valign.clear = valign.baseline = attr.clearclass(valign)


_textattrspreamble += "\\newdimen\\PyXDimenVShift%\n"

class _vshift(attr.sortbeforeattr, textattr):

    def __init__(self):
        attr.sortbeforeattr.__init__(self, [valign, parbox_pt, _localattr])

    def apply(self, expr):
        return r"%s\setbox0\hbox{{%s}}\lower\PyXDimenVShift\box0" % (self.setheightexpr(), expr)

class vshift(_vshift):
    "vertical down shift by a fraction of a character height"

    def __init__(self, lowerratio, heightstr="0"):
        _vshift.__init__(self)
        self.lowerratio = lowerratio
        self.heightstr = heightstr

    def setheightexpr(self):
        return r"\setbox0\hbox{{%s}}\PyXDimenVShift=%.5f\ht0" % (self.heightstr, self.lowerratio)

class _vshiftmathaxis(_vshift):
    "vertical down shift by the height of the math axis"

    def setheightexpr(self):
        return r"\setbox0\hbox{$\vcenter{\vrule width0pt}$}\PyXDimenVShift=\ht0"

    def apply_trafo(self, textbox):
        return textbox.transform(trafo.translate_pt(0, -textbox.mathaxis_pt*unit.scale["x"]), keep_anchor=True)


vshift.bottomzero = vshift(0)
vshift.middlezero = vshift(0.5)
vshift.topzero = vshift(1)
vshift.mathaxis = _vshiftmathaxis()
vshift.clear = attr.clearclass(_vshift)


defaultsizelist = ["normalsize", "large", "Large", "LARGE", "huge", "Huge",
None, "tiny", "scriptsize", "footnotesize", "small"]

class size(attr.sortbeforeattr, textattr):
    "font size"

    def __init__(self, sizeindex=None, sizename=None, sizelist=defaultsizelist):
        if (sizeindex is None and sizename is None) or (sizeindex is not None and sizename is not None):
            raise ValueError("either specify sizeindex or sizename")
        attr.sortbeforeattr.__init__(self, [_mathmode, _vshift])
        if sizeindex is not None:
            if sizeindex >= 0 and sizeindex < sizelist.index(None):
                self.size = sizelist[sizeindex]
            elif sizeindex < 0 and sizeindex + len(sizelist) > sizelist.index(None):
                self.size = sizelist[sizeindex]
            else:
                raise IndexError("index out of sizelist range")
        else:
            self.size = sizename

    def apply(self, expr):
        return r"\%s{}%s" % (self.size, expr)

size.tiny = size(-4)
size.scriptsize = size.script = size(-3)
size.footnotesize = size.footnote = size(-2)
size.small = size(-1)
size.normalsize = size.normal = size(0)
size.large = size(1)
size.Large = size(2)
size.LARGE = size(3)
size.huge = size(4)
size.Huge = size(5)
size.clear = attr.clearclass(size)


###############################################################################
# textengines
###############################################################################


class MonitorOutput(threading.Thread):

    def __init__(self, name, output):
        """Deadlock-safe output stream reader and monitor.

        An instance of this class creates a thread to continously read lines
        from a stream. By that a deadlock due to a full pipe is prevented. In
        addition, the stream content can be monitored for containing a certain
        string (see :meth:`expect` and :meth:`wait`) and return all the
        collected output (see :meth:`read`).

        :param string name: name to be used while logging in :meth:`wait` and
            :meth:`done`
        :param file output: output stream

        """
        self.output = output
        self._expect = queue.Queue(1)
        self._received = threading.Event()
        self._output = queue.Queue()
        threading.Thread.__init__(self, name=name)
        self.daemon = True
        self.start()

    def expect(self, s):
        """Expect a string on a **single** line in the output.

        This method must be called **before** the output occurs, i.e. before
        the input is written to the TeX/LaTeX process.

        :param s: expected string or ``None`` if output is expected to become
            empty
        :type s: str or None

        """
        self._expect.put_nowait(s)

    def read(self):
        """Read all output collected since its previous call.

        The output reading should be synchronized by the :meth:`expect`
        and :meth:`wait` methods.

        :returns: collected output from the stream
        :rtype: str

        """
        l = []
        try:
            while True:
                l.append(self._output.get_nowait())
        except queue.Empty:
            pass
        return "".join(l).replace("\r\n", "\n").replace("\r", "\n")

    def _wait(self, waiter, checker):
        """Helper method to implement :meth:`wait` and :meth:`done`.

        Waits for an event using the *waiter* and *checker* functions while
        providing user feedback to the ``pyx``-logger using the warning level
        according to the ``wait`` and ``showwait`` from the ``text`` section of
        the pyx :mod:`config`.

        :param function waiter: callback to wait for (the function gets called
            with a timeout parameter)
        :param function checker: callback returing ``True`` if
            waiting was successful
        :returns: ``True`` when wait was successful
        :rtype: bool

        """
        wait = config.getint("text", "wait", 60)
        showwait = config.getint("text", "showwait", 5)
        if showwait:
            waited = 0
            hasevent = False
            while waited < wait and not hasevent:
                if wait - waited > showwait:
                    waiter(showwait)
                    waited += showwait
                else:
                    waiter(wait - waited)
                    waited += wait - waited
                hasevent = checker()
                if not hasevent:
                    if waited < wait:
                        logger.warning("Still waiting for {} "
                                       "after {} (of {}) seconds..."
                                       .format(self.name, waited, wait))
                    else:
                        logger.warning("The timeout of {} seconds expired "
                                       "and {} did not respond."
                                       .format(waited, self.name))
            return hasevent
        else:
            waiter(wait)
            return checker()

    def wait(self):
        """Wait for the expected output to happen.

        Waits either until a line containing the string set by the previous
        :meth:`expect` call is found, or a timeout occurs.

        :returns: ``True`` when the expected string was found
        :rtype: bool

        """
        r = self._wait(self._received.wait, self._received.is_set)
        if r:
            self._received.clear()
        return r

    def done(self):
        """Waits until the output becomes empty.

        Waits either until the output becomes empty, or a timeout occurs.
        The generated output can still be catched by :meth:`read` after
        :meth:`done` was successful.

        In the proper workflow :meth:`expect` should be called with ``None``
        before the output completes, as otherwise a ``ValueError`` is raised
        in the :meth:`run`.

        :returns: ``True`` when the output has become empty
        :rtype: bool

        """
        return self._wait(self.join, lambda self=self: not self.is_alive())

    def _readline(self):
        """Read a line from the output.

        To be used **inside** the thread routine only.

        :returns: one line of the output as a string
        :rtype: str

        """
        while True:
            try:
                return self.output.readline()
            except IOError as e:
                if e.errno != errno.EINTR:
                     raise

    def run(self):
        """Thread routine.

        **Not** to be called from outside.

        :raises ValueError: output becomes empty while some string is expected

        """
        expect = None
        while True:
            line = self._readline()
            if expect is None:
                try:
                    expect = self._expect.get_nowait()
                except queue.Empty:
                    pass
            if not line:
                break
            self._output.put(line)
            if expect is not None:
                found = line.find(expect)
                if found != -1:
                    self._received.set()
                    expect = None
        self.output.close()
        if expect is not None:
            raise ValueError("{} finished unexpectedly".format(self.name))


class textbox_pt(box.rect, baseclasses.canvasitem): pass


class textextbox_pt(textbox_pt):

    def __init__(self, x_pt, y_pt, left_pt, right_pt, height_pt, depth_pt, do_finish, fontmap, singlecharmode, fillstyles):
        """Text output.

        An instance of this class contains the text output generated by PyX. It
        is a :class:`baseclasses.canvasitem` and thus can be inserted into a
        canvas.

        .. A text has a center (x_pt, y_pt) as well as extents in x-direction
        .. (left_pt and right_pt) and y-direction (hight_pt and depth_pt). The
        .. textbox positions the output accordingly and scales it by the
        .. x-scale from the :mod:`unit`.

        .. :param float x_pt: x coordinate of the center in pts
        .. :param float y_pt: y coordinate of the center in pts
        .. :param float left_pt: unscaled left extent in pts
        .. :param float right_pt: unscaled right extent in pts
        .. :param float height_pt: unscaled height in pts
        .. :param float depth_pt: unscaled depth in pts
        .. :param function do_finish: callable to execute :meth:`readdvipage`
        .. :param fontmap: force a fontmap to be used (instead of the default
        ..     depending on the output format)
        .. :type fontmap: None or fontmap
        .. :param bool singlecharmode: position each character separately
        .. :param fillstyles: fill styles to be applied
        .. :type fillstyles: list of fillstyles

        """
        self.left = left_pt*unit.x_pt       #: left extent of the text (PyX length)
        self.right = right_pt*unit.x_pt     #: right extent of the text (PyX length)
        self.width = self.left + self.right #: width of the text (PyX length)
        self.height = height_pt*unit.x_pt   #: height of the text (PyX length)
        self.depth = depth_pt*unit.x_pt     #: height of the text (PyX length)

        self.do_finish = do_finish
        self.fontmap = fontmap
        self.singlecharmode = singlecharmode
        self.fillstyles = fillstyles

        self.texttrafo = trafo.scale(unit.scale["x"]).translated_pt(x_pt, y_pt)
        box.rect_pt.__init__(self, x_pt - left_pt*unit.scale["x"], y_pt - depth_pt*unit.scale["x"],
                                   (left_pt + right_pt)*unit.scale["x"],
                                   (depth_pt + height_pt)*unit.scale["x"],
                                   abscenter_pt = (left_pt*unit.scale["x"], depth_pt*unit.scale["x"]))

        self._dvicanvas = None

    def transform(self, *trafos, keep_anchor=False):
        box.rect.transform(self, *trafos, keep_anchor=keep_anchor)
        for trafo in trafos:
            self.texttrafo = trafo * self.texttrafo
        if self._dvicanvas is not None:
            for trafo in trafos:
                self._dvicanvas.trafo = trafo * self._dvicanvas.trafo

    def readdvipage(self, dvifile, page):
        self._dvicanvas = dvifile.readpage([ord("P"), ord("y"), ord("X"), page, 0, 0, 0, 0, 0, 0],
                                           fontmap=self.fontmap, singlecharmode=self.singlecharmode, attrs=[self.texttrafo] + self.fillstyles)

    @property
    def dvicanvas(self):
        if self._dvicanvas is None:
            self.do_finish()
        return self._dvicanvas

    def marker(self, name):
        """Return the position of a marker.

        :param str name: name of the marker
        :returns: position of the marker
        :rtype: tuple of two PyX lengths

        This method returns the position of the marker of the given name
        within, previously defined by the ``\\PyXMarker{name}`` macro in the
        typeset text. Please do not use the ``@`` character within your name to
        prevent name clashes with PyX internal uses (although we don’t the
        marker feature internally right now).

        Similar to generating actual output, the marker function accesses the
        DVI output, stopping. The :ref:`texipc` feature will allow for this access
        without stopping the TeX interpreter.

        """
        return self.texttrafo.apply(*self.dvicanvas.markers[name])

    def textpath(self):
        textpath = path.path()
        for item in self.dvicanvas.items:
            textpath += item.textpath()
        return textpath.transformed(self.texttrafo)

    def processPS(self, file, writer, context, registry, bbox):
        abbox = bboxmodule.empty()
        self.dvicanvas.processPS(file, writer, context, registry, abbox)
        bbox += box.rect.bbox(self)

    def processPDF(self, file, writer, context, registry, bbox):
        abbox = bboxmodule.empty()
        self.dvicanvas.processPDF(file, writer, context, registry, abbox)
        bbox += box.rect.bbox(self)

    def processSVG(self, xml, writer, context, registry, bbox):
        abbox = bboxmodule.empty()
        self.dvicanvas.processSVG(xml, writer, context, registry, abbox)
        bbox += box.rect.bbox(self)


class _marker:
    pass


class errordetail:
    "Constants defining the verbosity of the :exc:`TexResultError`."
    none = 0    #: Without any input and output.
    default = 1 #: Input and parsed output shortend to 5 lines.
    full = 2    #: Full input and unparsed as well as parsed output.


class Tee(object):

    def __init__(self, *files):
        """Apply write, flush, and close to each of the given files."""
        self.files = files

    def write(self, data):
        for file in self.files:
            file.write(data)

    def flush(self):
        for file in self.files:
            file.flush()

    def close(self):
        for file in self.files:
            file.close()

# The tex engine state represents the next (or current) execute state.
STATE_START, STATE_PREAMBLE, STATE_TYPESET, STATE_DONE = range(4)
PyXBoxPattern = re.compile(r"PyXBox:page=(?P<page>\d+),lt=(?P<lt>-?\d*((\d\.?)|(\.?\d))\d*)pt,rt=(?P<rt>-?\d*((\d\.?)|(\.?\d))\d*)pt,ht=(?P<ht>-?\d*((\d\.?)|(\.?\d))\d*)pt,dp=(?P<dp>-?\d*((\d\.?)|(\.?\d))\d*)pt:")
dvi_pattern = re.compile(r'Output written on .*texput\.dvi"? \((?P<page>\d+) pages?, \d+ bytes\)\.', re.DOTALL)

class TexDoneError(Exception):
    pass


class SingleEngine:

    #: default :class:`texmessage` parsers at interpreter startup
    texmessages_start_default = [texmessage.start]
    #: default :class:`texmessage` parsers at interpreter shutdown
    texmessages_end_default = [texmessage.end, texmessage.font_warning, texmessage.rerun_warning, texmessage.nobbl_warning]
    #: default :class:`texmessage` parsers for preamble output
    texmessages_preamble_default = [texmessage.load]
    #: default :class:`texmessage` parsers for typeset output
    texmessages_run_default = [texmessage.font_warning, texmessage.box_warning, texmessage.package_warning,
                                      texmessage.load_def, texmessage.load_graphics]

    def __init__(self, cmd,
                       texenc="ascii",
                       usefiles=[],
                       texipc=config.getboolean("text", "texipc", 0),
                       copyinput=None,
                       dvitype=False,
                       errordetail=errordetail.default,
                       texmessages_start=[],
                       texmessages_end=[],
                       texmessages_preamble=[],
                       texmessages_run=[]):
        """Base class for the TeX interface.

        .. note:: This class cannot be used directly. It is the base class for
                  all tex engines and provides most of the implementation.
                  Still, to the end user the parameters except for *cmd*
                  are important, as they are preserved in derived classes
                  usually.

        :param cmd: command and arguments to start the TeX interpreter
        :type cmd: list of str
        :param str texenc: encoding to use in the communication with the TeX
            interpreter
        :param usefiles: list of supplementary files to be copied to and from
            the temporary working directory (see :ref:`debug` for usage
            details)
        :type usefiles: list of str
        :param bool texipc: :ref:`texipc` flag.
        :param copyinput: filename or file to be used to store a copy of all
            the input passed to the TeX interpreter
        :type copyinput: None or str or file
        :param bool dvitype: flag to turn on dvitype-like output
        :param errordetail: verbosity of the :exc:`TexResultError`
        :type errordetail: :class:`errordetail`
        :param texmessages_start: additional message parsers at interpreter
            startup
        :type texmessages_start: list of :class:`texmessage` parsers
        :param texmessages_end: additional message parsers at interpreter
            shutdown
        :type texmessages_end: list of :class:`texmessage` parsers
        :param texmessages_preamble: additional message parsers for preamble
            output
        :type texmessages_preamble: list of :class:`texmessage` parsers
        :param texmessages_run: additional message parsers for typset output
        :type texmessages_run: list of :class:`texmessage` parsers

        """
        self.cmd = cmd
        self.texenc = texenc
        self.usefiles = usefiles
        self.texipc = texipc
        self.copyinput = copyinput
        self.dvitype = dvitype
        self.errordetail = errordetail
        self.texmessages_start = texmessages_start
        self.texmessages_end = texmessages_end
        self.texmessages_preamble = texmessages_preamble
        self.texmessages_run = texmessages_run

        self.state = STATE_START
        self.executeid = 0
        self.page = 0

        self.needdvitextboxes = [] # when texipc-mode off
        self.dvifile = None

    def _cleanup(self):
        """Clean-up TeX interpreter and tmp directory.

        This funtion is hooked up in atexit to quit the TeX interpreter, to
        save the contents of usefiles, and to remove the temporary directory.

        """
        try:
            if self.state > STATE_START:
                if self.state < STATE_DONE:
                    self.do_finish(cleanup=False)
                    if self.state < STATE_DONE: # cleanup while TeX is still running?
                        self.texoutput.expect(None)
                        self.force_done()
                        for f, msg in [(self.texinput.close, "We tried to communicate to %s to quit itself, but this seem to have failed. Trying by terminate signal now ...".format(self.name)),
                                       (self.popen.terminate, "Failed, too. Trying by kill signal now ..."),
                                       (self.popen.kill, "We tried very hard to get rid of the %s process, but we ultimately failed (as far as we can tell). Sorry.".format(self.name))]:
                            f()
                            if self.texoutput.done():
                                break
                            logger.warning(msg)
                self.popen.wait()
                for usefile in self.usefiles:
                    extpos = usefile.rfind(".")
                    try:
                        shutil.move(os.path.join(self.tmpdir, "texput" + usefile[extpos:]), usefile)
                    except EnvironmentError:
                        logger.warning("Could not save '{}'.".format(usefile))
                        if os.path.isfile(usefile):
                            try:
                                os.unlink(usefile)
                            except EnvironmentError:
                                logger.warning("Failed to remove spurious file '{}'.".format(usefile))
        finally:
            shutil.rmtree(self.tmpdir, ignore_errors=True)

    def _execute(self, expr, texmessages, oldstate, newstate):
        """Execute TeX expression.

        :param str expr: expression to be passed to TeX
        :param texmessages: message parsers to analyse the textual output of
            TeX
        :type texmessages: list of :class:`texmessage` parsers
        :param int oldstate: state of the TeX interpreter prior to the
            expression execution
        :param int newstate: state of the TeX interpreter after to the
            expression execution

        """
        assert STATE_PREAMBLE <= oldstate <= STATE_TYPESET
        assert oldstate == self.state
        assert newstate >= oldstate
        if newstate == STATE_DONE:
            self.texoutput.expect(None)
            self.texinput.write(expr)
        else:

            # test to encode expr early to not pile up expected results
            # if the expression won't make it to the texinput at all
            # (which would otherwise harm a proper cleanup)
            expr.encode(self.texenc)

            if oldstate == newstate == STATE_TYPESET:
                self.page += 1
                expr = "\\ProcessPyXBox{%s%%\n}{%i}" % (expr, self.page)
            self.executeid += 1
            self.texoutput.expect("PyXInputMarker:executeid=%i:" % self.executeid)
            expr += "%%\n\\PyXInput{%i}%%\n" % self.executeid
            self.texinput.write(expr)
        self.texinput.flush()
        self.state = newstate
        if newstate == STATE_DONE:
            wait_ok = self.texoutput.done()
        else:
            wait_ok = self.texoutput.wait()
        try:
            parsed = unparsed = self.texoutput.read()
            if not wait_ok:
                raise TexResultError("TeX didn't respond as expected within the timeout period.")
            if newstate != STATE_DONE:
                parsed, m = remove_string("PyXInputMarker:executeid=%s:" % self.executeid, parsed)
                if not m:
                    raise TexResultError("PyXInputMarker expected")
                if oldstate == newstate == STATE_TYPESET:
                    parsed, m = remove_pattern(PyXBoxPattern, parsed, ignore_nl=False)
                    if not m:
                        raise TexResultError("PyXBox expected")
                    if m.group("page") != str(self.page):
                        raise TexResultError("Wrong page number in PyXBox")
                    extent_pt = [float(x)*72/72.27 for x in m.group("lt", "rt", "ht", "dp")]
                    parsed, m = remove_string("[80.121.88.%s]" % self.page, parsed)
                    if not m:
                        raise TexResultError("PyXPageOutMarker expected")
            else:
                # check for "Output written on ...dvi (1 page, 220 bytes)."
                if self.page:
                    parsed, m = remove_pattern(dvi_pattern, parsed)
                    if not m:
                        raise TexResultError("TeX dvifile messages expected")
                    if m.group("page") != str(self.page):
                        raise TexResultError("wrong number of pages reported")
                else:
                    parsed, m = remove_string("No pages of output.", parsed)
                    if not m:
                        raise TexResultError("no dvifile expected")

            for t in texmessages:
                parsed = t(parsed)
            if parsed.replace(r"(Please type a command or say `\end')", "").replace(" ", "").replace("*\n", "").replace("\n", ""):
                raise TexResultError("unhandled TeX response (might be an error)")
        except TexResultError as e:
            if self.errordetail > errordetail.none:
                def add(msg): e.args = (e.args[0] + msg,)
                add("\nThe expression passed to TeX was:\n{}".format(indent_text(expr.rstrip())))
                if self.errordetail == errordetail.full:
                    add("\nThe return message from TeX was:\n{}".format(indent_text(unparsed.rstrip())))
                if self.errordetail == errordetail.default:
                    if parsed.count('\n') > 6:
                        parsed = "\n".join(parsed.split("\n")[:5] + ["(cut after 5 lines; use errordetail.full for all output)"])
                add("\nAfter parsing the return message from TeX, the following was left:\n{}".format(indent_text(parsed.rstrip())))
            raise e
        if oldstate == newstate == STATE_TYPESET:
            return extent_pt

    def do_start(self):
        """Setup environment and start TeX interpreter."""
        assert self.state == STATE_START

        chroot = config.get("text", "chroot", "")
        if chroot:
            chroot_tmpdir = config.get("text", "tmpdir", "/tmp")
            chroot_tmpdir_rel = os.path.relpath(chroot_tmpdir, os.sep)
            base_tmpdir = os.path.join(chroot, chroot_tmpdir_rel)
        else:
            base_tmpdir = config.get("text", "tmpdir", None)
        self.tmpdir = tempfile.mkdtemp(prefix="pyx", dir=base_tmpdir)
        atexit.register(self._cleanup)
        for usefile in self.usefiles:
            extpos = usefile.rfind(".")
            try:
                os.rename(usefile, os.path.join(self.tmpdir, "texput" + usefile[extpos:]))
            except OSError:
                pass
        if chroot:
            tex_tmpdir = os.sep + os.path.relpath(self.tmpdir, chroot)
        else:
            tex_tmpdir = self.tmpdir
        cmd = self.cmd + ['--output-directory', tex_tmpdir]
        if self.texipc:
            cmd.append("--ipc")
        self.popen = config.Popen(cmd, stdin=config.PIPE, stdout=config.PIPE, stderr=config.STDOUT, bufsize=0)
        self.texinput = io.TextIOWrapper(self.popen.stdin, encoding=self.texenc)
        if self.copyinput:
            try:
                self.copyinput.write
            except AttributeError:
                self.texinput = Tee(open(self.copyinput, "w", encoding=self.texenc), self.texinput)
            else:
                self.texinput = Tee(self.copyinput, self.texinput)
        self.texoutput = MonitorOutput(self.name, io.TextIOWrapper(self.popen.stdout, encoding=self.texenc, errors="surrogateescape"))

        self.state = STATE_PREAMBLE
        self._execute("\\scrollmode\n\\raiseerror%\n" # switch to and check scrollmode
                      "\\def\\PyX{P\\kern-.3em\\lower.5ex\\hbox{Y}\\kern-.18em X}%\n" # just the PyX Logo
                      "\\gdef\\PyXBoxHAlign{0}%\n" # global PyXBoxHAlign (0.0-1.0) for the horizontal alignment, default to 0
                      "\\newbox\\PyXBox%\n" # PyXBox will contain the output
                      "\\newbox\\PyXBoxHAligned%\n" # PyXBox will contain the horizontal aligned output
                      "\\newdimen\\PyXDimenHAlignLT%\n" # PyXDimenHAlignLT/RT will contain the left/right extent
                      "\\newdimen\\PyXDimenHAlignRT%\n" +
                      _textattrspreamble + # insert preambles for textattrs macros
                      "\\long\\def\\ProcessPyXBox#1#2{%\n" # the ProcessPyXBox definition (#1 is expr, #2 is page number)
                      "\\setbox\\PyXBox=\\hbox{{#1}}%\n" # push expression into PyXBox
                      "\\PyXDimenHAlignLT=\\PyXBoxHAlign\\wd\\PyXBox%\n" # calculate the left/right extent
                      "\\PyXDimenHAlignRT=\\wd\\PyXBox%\n"
                      "\\advance\\PyXDimenHAlignRT by -\\PyXDimenHAlignLT%\n"
                      "\\gdef\\PyXBoxHAlign{0}%\n" # reset the PyXBoxHAlign to the default 0
                      "\\immediate\\write16{PyXBox:page=#2," # write page and extents of this box to stdout
                                                  "lt=\\the\\PyXDimenHAlignLT,"
                                                  "rt=\\the\\PyXDimenHAlignRT,"
                                                  "ht=\\the\\ht\\PyXBox,"
                                                  "dp=\\the\\dp\\PyXBox:}%\n"
                      "\\ht\\PyXBox0pt%\n" # baseline alignment (hight to zero)
                      "\\setbox\\PyXBoxHAligned=\\hbox{\\kern-\\PyXDimenHAlignLT\\box\\PyXBox}%\n" # align horizontally
                      "{\\count0=80\\count1=121\\count2=88\\count3=#2\\shipout\\box\\PyXBoxHAligned}}%\n" # shipout PyXBox to Page 80.121.88.<page number>
                      "\\def\\PyXInput#1{\\immediate\\write16{PyXInputMarker:executeid=#1:}}%\n" # write PyXInputMarker to stdout
                      "\\def\\PyXMarker#1{\\hskip0pt\\special{PyX:marker #1}}%", # write PyXMarker special into the dvi-file
                      self.texmessages_start_default + self.texmessages_start, STATE_PREAMBLE, STATE_PREAMBLE)

    def do_preamble(self, expr, texmessages):
        """Ensure preamble mode and execute expr."""
        if self.state < STATE_PREAMBLE:
            self.do_start()
        self._execute(expr, texmessages, STATE_PREAMBLE, STATE_PREAMBLE)

    def do_typeset(self, expr, texmessages):
        """Ensure typeset mode and typeset expr."""
        if self.state < STATE_PREAMBLE:
            self.do_start()
        if self.state < STATE_TYPESET:
            self.go_typeset()
        return self._execute(expr, texmessages, STATE_TYPESET, STATE_TYPESET)

    def do_finish(self, cleanup=True):
        """Teardown TeX interpreter and cleanup environment.

        :param bool cleanup: use _cleanup regularly/explicitly (not via atexit)
        """
        if self.state == STATE_DONE:
            return
        if self.state < STATE_TYPESET:
            self.go_typeset()
        self.go_finish()
        assert self.state == STATE_DONE
        self.texinput.close()            # close the input queue and
        self.texoutput.done()            # wait for finish of the output

        if self.needdvitextboxes:
            dvifilename = os.path.join(self.tmpdir, "texput.dvi")
            self.dvifile = dvifile.DVIfile(dvifilename, debug=self.dvitype)
            page = 1
            for box in self.needdvitextboxes:
                box.readdvipage(self.dvifile, page)
                page += 1
        if self.dvifile is not None and self.dvifile.readpage(None) is not None:
            raise ValueError("end of dvifile expected but further pages follow")
        if cleanup:
            atexit.unregister(self._cleanup)
            self._cleanup()

    def preamble(self, expr, texmessages=[]):
        """Execute a preamble.

        :param str expr: expression to be executed
        :param texmessages: additional message parsers
        :type texmessages: list of :class:`texmessage` parsers

        Preambles must not generate output, but are used to load files, perform
        settings, define macros, *etc*. In LaTeX mode, preambles are executed
        before ``\\begin{document}``. The method can be called multiple times,
        but only prior to :meth:`SingleEngine.text` and
        :meth:`SingleEngine.text_pt`.

        """
        texmessages = self.texmessages_preamble_default + self.texmessages_preamble + texmessages
        self.do_preamble(expr, texmessages)

    def text_pt(self, x_pt, y_pt, expr, textattrs=[], texmessages=[], fontmap=None, singlecharmode=False):
        """Typeset text.

        :param float x_pt: x position in pts
        :param float y_pt: y position in pts
        :param expr: text to be typeset
        :type expr: str or :class:`MultiEngineText`
        :param textattrs: styles and attributes to be applied to the text
        :type textattrs: list of  :class:`textattr, :class:`trafo.trafo_pt`,
            and :class:`style.fillstyle`
        :param texmessages: additional message parsers
        :type texmessages: list of :class:`texmessage` parsers
        :param fontmap: force a fontmap to be used (instead of the default
            depending on the output format)
        :type fontmap: None or fontmap
        :param bool singlecharmode: position each character separately
        :returns: text output insertable into a canvas.
        :rtype: :class:`textextbox_pt`
        :raises: :exc:`TexDoneError`: when the TeX interpreter has been
            terminated already.

        """
        if self.state == STATE_DONE:
            raise TexDoneError("typesetting process was terminated already")
        textattrs = attr.mergeattrs(textattrs) # perform cleans
        attr.checkattrs(textattrs, [textattr, trafo.trafo_pt, style.fillstyle])
        trafos = attr.getattrs(textattrs, [trafo.trafo_pt])
        fillstyles = attr.getattrs(textattrs, [style.fillstyle])
        textattrs = attr.getattrs(textattrs, [textattr])
        if isinstance(expr, MultiEngineText):
            expr = expr.tex
        for ta in textattrs[::-1]:
            expr = ta.apply(expr)
        first = self.state < STATE_TYPESET
        left_pt, right_pt, height_pt, depth_pt = self.do_typeset(expr, self.texmessages_run_default + self.texmessages_run + texmessages)
        if self.texipc and first:
            self.dvifile = dvifile.DVIfile(os.path.join(self.tmpdir, "texput.dvi"), debug=self.dvitype)
        box = textextbox_pt(x_pt, y_pt, left_pt, right_pt, height_pt, depth_pt, self.do_finish, fontmap, singlecharmode, fillstyles)
        for t in trafos:
            box.reltransform(t) # TODO: should trafos really use reltransform???
                                #       this is quite different from what we do elsewhere!!!
                                #       see https://sourceforge.net/mailarchive/forum.php?thread_id=9137692&forum_id=23700
        if self.texipc:
            box.readdvipage(self.dvifile, self.page)
        else:
            self.needdvitextboxes.append(box)
        return box

    def text(self, x, y, *args, **kwargs):
        """Typeset text.

        This method is identical to :meth:`text_pt` with the only difference of
        using PyX lengths to position the output.

        :param x: x position
        :type x: PyX length
        :param y: y position
        :type y: PyX length

        """
        return self.text_pt(unit.topt(x), unit.topt(y), *args, **kwargs)


class SingleTexEngine(SingleEngine):

    def __init__(self, cmd=config.getlist("text", "tex", ["tex"]), lfs="10pt", **kwargs):
        """Plain TeX interface.

        This class adjusts the :class:`SingleEngine` to use plain TeX.

        :param cmd: command and arguments to start the TeX interpreter
        :type cmd: list of str
        :param lfs: resemble LaTeX font settings within plain TeX by loading a
            lfs-file
        :type lfs: str or None
        :param kwargs: additional arguments passed to :class:`SingleEngine`

        An lfs-file is a file defining a set of font commands like ``\\normalsize``
        by font selection commands in plain TeX. Several of those files
        resembling standard settings of LaTeX are distributed along with PyX in
        the ``pyx/data/lfs`` directory. This directory also contains a LaTeX
        file to create lfs files for different settings (LaTeX class, class
        options, and style files).

        """
        super().__init__(cmd=cmd, **kwargs)
        self.lfs = lfs
        self.name = "TeX"

    def go_typeset(self):
        assert self.state == STATE_PREAMBLE
        self.state = STATE_TYPESET

    def go_finish(self):
        self._execute("\\end%\n", self.texmessages_end_default + self.texmessages_end, STATE_TYPESET, STATE_DONE)

    def force_done(self):
        self.texinput.write("\n\\end\n")

    def do_start(self):
        super().do_start()
        if self.lfs:
            if not self.lfs.endswith(".lfs"):
                self.lfs = "%s.lfs" % self.lfs
            with config.open(self.lfs, [config.format.pyx]) as lfsfile:
                lfsdef = lfsfile.read().decode("ascii")
            self._execute(lfsdef, [], STATE_PREAMBLE, STATE_PREAMBLE)
            self._execute("\\normalsize%\n", [], STATE_PREAMBLE, STATE_PREAMBLE)
        self._execute("\\newdimen\\linewidth\\newdimen\\textwidth%\n", [], STATE_PREAMBLE, STATE_PREAMBLE)


class SingleLatexEngine(SingleEngine):

    #: default :class:`texmessage` parsers at LaTeX class loading
    texmessages_docclass_default = [texmessage.load]
    #: default :class:`texmessage` parsers at ``\begin{document}``
    texmessages_begindoc_default = [texmessage.load, texmessage.no_aux]

    def __init__(self, cmd=config.getlist("text", "latex", ["latex"]),
                       docclass="article", docopt=None, pyxgraphics=True,
                       texmessages_docclass=[], texmessages_begindoc=[], **kwargs):
        """LaTeX interface.

        This class adjusts the :class:`SingleEngine` to use LaTeX.

        :param cmd: command and arguments to start the TeX interpreter
            in LaTeX mode
        :type cmd: list of str
        :param str docclass: document class
        :param docopt: document loading options
        :type docopt: str or None
        :param bool pyxgraphics: activate graphics bundle support, see
            :ref:`pyxgraphics`
        :param texmessages_docclass: additional message parsers at LaTeX class
            loading
        :type texmessages_docclass: list of :class:`texmessage` parsers
        :param texmessages_begindoc: additional message parsers at
            ``\\begin{document}``
        :type texmessages_begindoc: list of :class:`texmessage` parsers
        :param kwargs: additional arguments passed to :class:`SingleEngine`

        """
        super().__init__(cmd=cmd, **kwargs)
        self.docclass = docclass
        self.docopt = docopt
        self.pyxgraphics = pyxgraphics
        self.texmessages_docclass = texmessages_docclass
        self.texmessages_begindoc = texmessages_begindoc
        self.name = "LaTeX"

    def go_typeset(self):
        self._execute("\\begin{document}", self.texmessages_begindoc_default + self.texmessages_begindoc, STATE_PREAMBLE, STATE_TYPESET)

    def go_finish(self):
        self._execute("\\end{document}%\n", self.texmessages_end_default + self.texmessages_end, STATE_TYPESET, STATE_DONE)

    def force_done(self):
        self.texinput.write("\n\\catcode`\\@11\\relax\\@@end\n")

    def do_start(self):
        super().do_start()
        if self.pyxgraphics:
            with config.open("pyx.def", [config.format.pyx]) as source, open(os.path.join(self.tmpdir, "pyx.def"), "wb") as dest:
                dest.write(source.read())
            self._execute("\\makeatletter%\n"
                          "\\let\\saveProcessOptions=\\ProcessOptions%\n"
                          "\\def\\ProcessOptions{%\n"
                          "\\def\\Gin@driver{" + self.tmpdir.replace(os.sep, "/") + "/pyx.def}%\n"
                          "\\def\\c@lor@namefile{dvipsnam.def}%\n"
                          "\\saveProcessOptions}%\n"
                          "\\makeatother",
                          [], STATE_PREAMBLE, STATE_PREAMBLE)
        if self.docopt is not None:
            self._execute("\\documentclass[%s]{%s}" % (self.docopt, self.docclass),
                          self.texmessages_docclass_default + self.texmessages_docclass, STATE_PREAMBLE, STATE_PREAMBLE)
        else:
            self._execute("\\documentclass{%s}" % self.docclass,
                          self.texmessages_docclass_default + self.texmessages_docclass, STATE_PREAMBLE, STATE_PREAMBLE)


def reset_for_tex_done(f):
    @functools.wraps(f)
    def wrapped(self, *args, **kwargs):
        try:
            return f(self, *args, **kwargs)
        except TexDoneError:
            self.reset(reinit=True)
            return f(self, *args, **kwargs)
    return wrapped


class MultiEngine:

    def __init__(self, cls, *args, **kwargs):
        """A restartable :class:`SingleEngine` class

        :param cls: the class being wrapped
        :type cls: :class:`SingleEngine` class
        :param list args: args at class instantiation
        :param dict kwargs: keyword args at at class instantiation

        """
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.reset()

    def preamble(self, expr, texmessages=[]):
        "resembles :meth:`SingleEngine.preamble`"
        self.preambles.append((expr, texmessages))
        self.instance.preamble(expr, texmessages)

    @reset_for_tex_done
    def text_pt(self, *args, **kwargs):
        "resembles :meth:`SingleEngine.text_pt`"
        return self.instance.text_pt(*args, **kwargs)

    @reset_for_tex_done
    def text(self, *args, **kwargs):
        "resembles :meth:`SingleEngine.text`"
        return self.instance.text(*args, **kwargs)

    def reset(self, reinit=False):
        """Start a new :class:`SingleEngine` instance

        :param bool reinit: replay :meth:`preamble` calls on the new instance

        After executing this function further preamble calls are allowed,
        whereas once a text output has been created, :meth:`preamble` calls are
        forbidden.

        """
        self.instance = self.cls(*self.args, **self.kwargs)
        if reinit:
            for expr, texmessages in self.preambles:
                self.instance.preamble(expr, texmessages)
        else:
            self.preambles = []


class TexEngine(MultiEngine):

    def __init__(self, *args, **kwargs):
        """A restartable :class:`SingleTexEngine` class

        :param list args: args at class instantiation
        :param dict kwargs: keyword args at at class instantiation

        """
        super().__init__(SingleTexEngine, *args, **kwargs)


class LatexEngine(MultiEngine):

    def __init__(self, *args, **kwargs):
        """A restartable :class:`SingleLatexEngine` class

        :param list args: args at class instantiation
        :param dict kwargs: keyword args at at class instantiation

        """
        super().__init__(SingleLatexEngine, *args, **kwargs)


from pyx import deco
from pyx.font import T1font
from pyx.font.t1file import T1File
from pyx.font.afmfile import AFMfile
from pyx.font.pfmfile import PFMfile


class MultiEngineText:

    def __init__(self, tex, unicode):
        """Container for text to be typeset for both, TeX based engines and the
        UnicodeEngine.

        :param str tex: text for TeX based engines
        :param unicode: text for the :class:`UnicodeEngine`
        :type unicode: str, :class:`Text`, or :class:`StackedText or a list of those

        """
        self.tex = tex
        self.unicode = unicode


class Text:

    def __init__(self, text, scale=1, shift=0):
        """Text for the UnicodeEngine with basic typesetting features

        :param str text: text to be typeset
        :param float scale: scale
        :param float shift: vertical shift in units of the text size (without the scale)

        """
        self.text = text
        self.scale = scale
        self.shift = shift


class StackedText:

    def __init__(self, texts, frac=False, align=0):
        """Stack text above each other for the UnicodeEngine

        :param list texts: texts to be typeset above each other
        :param bool frac: add a fractional line (for two texts only)
        :param float align: horizontal alignment of the text where
            0 is left, 0.5 is centered, and 1 is right

        """
        assert not frac or len(texts) == 2
        self.texts = texts
        self.frac = frac
        self.align = align


class unicodetextbox_pt(textbox_pt):

    def __init__(self, x_pt, y_pt, texts, font, size, mathmode=False):
        self.font = font
        self.size = size
        self.canvas = canvas.canvas()
        self.texttrafo = trafo.scale(unit.scale["x"]).translated_pt(x_pt, y_pt)

        if isinstance(texts, (str, Text, StackedText)):
            texts = [texts]

        x_pt = 0
        for text in texts:
            if isinstance(text, str):
                text = Text(text)
            if isinstance(text, Text):
                if mathmode:
                    text_fragments = text.text.split('-')
                else:
                    text_fragments = [text.text]
                for i, text_fragment in enumerate(text_fragments):
                    if i:
                        self.canvas.fill(path.rect_pt(x_pt+0.5*(self.minuswidth_pt-self.minuslength_pt), self.mathaxis_pt-0.5*self.minusthickness_pt, self.minuslength_pt, self.minusthickness_pt))
                        x_pt += self.minuswidth_pt
                    if text_fragment:
                        try:
                            t = self.font.text_pt(x_pt, text.shift*self.size, text_fragment, text.scale*self.size)
                            x_pt += t.bbox().width_pt()
                        except Exception:
                            assert '·' in text_fragment
                            t = self.font.text_pt(x_pt, text.shift*self.size, text_fragment.replace('·', 'x'), text.scale*self.size)
                            x_pt += t.bbox().width_pt()
                        self.canvas.insert(t)
            else:
                assert isinstance(text, StackedText)
                shift = self.mathaxis_pt if text.frac else 0
                ts = [self.font.text_pt(x_pt, text.shift*self.size+shift, text.text, text.scale*self.size)
                      for text in text.texts]
                width_pt = max(t.bbox().width_pt() for t in ts)
                if text.frac:
                    self.canvas.fill(path.rect_pt(x_pt, self.mathaxis_pt-0.5*self.minusthickness_pt, width_pt, self.minusthickness_pt))
                for t in ts:
                    self.canvas.insert(t, [trafo.translate_pt(text.align*(width_pt-t.bbox().width_pt()), 0)] if text.align else [])
                x_pt += width_pt

        bbox = self.canvas.bbox()
        bbox.includepoint_pt(0, 0)
        bbox.includepoint_pt(x_pt, 0)
        box.rect_pt.__init__(self, bbox.llx_pt, bbox.lly_pt, bbox.urx_pt-bbox.llx_pt, bbox.ury_pt-bbox.lly_pt, abscenter_pt = (0, 0))
        box.rect.transform(self, self.texttrafo)

    def _extract_minus_properties(self):
        minus_path = self.font.text_pt(0, 0, '=', 1).textpath().normpath()
        minus_path.normsubpaths = [normsubpath for normsubpath in minus_path.normsubpaths if normsubpath]
        self.font.minusthickness_pt = max(normsubpath.bbox().height_pt() for normsubpath in minus_path.normsubpaths)
        self.font.halfminuswidth_pt, self.font.mathaxis_pt = minus_path.bbox().center_pt()
        self.font.minuslength_pt = minus_path.bbox().width_pt()

    @property
    def mathaxis_pt(self):
        if not hasattr(self.font, "mathaxis_pt"):
            self._extract_minus_properties()
        return self.font.mathaxis_pt*self.size

    @property
    def minuswidth_pt(self):
        if not hasattr(self.font, "halfminuswidth_pt"):
            self._extract_minus_properties()
        return 2*self.font.halfminuswidth_pt*self.size

    @property
    def minuslength_pt(self):
        if not hasattr(self.font, "minuslength_pt"):
            self._extract_minus_properties()
        return self.font.minuslength_pt*self.size

    @property
    def minusthickness_pt(self):
        if not hasattr(self.font, "minusthickness_pt"):
            self._extract_minus_properties()
        return self.font.minusthickness_pt*self.size

    def transform(self, *trafos, keep_anchor=False):
        box.rect.transform(self, *trafos, keep_anchor=keep_anchor)
        for trafo in trafos:
            self.texttrafo = trafo * self.texttrafo

    def bbox(self):
        return self.canvas.bbox().transformed(self.texttrafo)

    def textpath(self):
        r = path.path()
        for item in self.canvas.items:
            if isinstance(item, canvas.canvas):
                for subitem in item.items:
                    r += subitem.textpath().transformed(item.trafo)
            elif isinstance(item, deco.decoratedpath):
                r += item.path
            else:
                r += item.textpath()
        return r.transformed(self.texttrafo)

    def processPS(self, file, writer, context, registry, bbox):
        c = canvas.canvas([self.texttrafo])
        c.insert(self.canvas)
        c.processPS(file, writer, context, registry, bbox)

    def processPDF(self, file, writer, context, registry, bbox):
        c = canvas.canvas([self.texttrafo])
        c.insert(self.canvas)
        c.processPDF(file, writer, context, registry, bbox)

    def processSVG(self, xml, writer, context, registry, bbox):
        c = canvas.canvas([self.texttrafo])
        c.insert(self.canvas)
        c.processSVG(xml, writer, context, registry, bbox)


class FontMetricType:

    #: :meta hide-value:
    #:
    #: use afm font metric
    afm = 1

    #: :meta hide-value:
    #:
    #: use pfm font metric
    pfm = 2


class UnicodeEngine:

    def __init__(self, fontname="cmr10", size=10, font_metric_type=FontMetricType.afm):
        with config.open(fontname, [config.format.type1]) as f:
            t1file = T1File.from_PF_bytes(f.read())
        if font_metric_type == FontMetricType.afm:
            with config.open(fontname, [config.format.afm], ascii=True) as f:
                metric_file = AFMfile(f)
        elif font_metric_type == FontMetricType.pfm:
            with config.open(fontname, [config.format.pfm]) as f:
                metric_file = PFMfile(f, t1file)
        else:
            raise ValueError("invalid value for matric")
        self.font = T1font(t1file, metric_file)
        self.size = size

    def preamble(self):
        raise NotImplemented()

    def reset(self):
        raise NotImplemented()

    def text_pt(self, x_pt, y_pt, text, textattrs=[], texmessages=[], fontmap=None, singlecharmode=False):

        textattrs = attr.mergeattrs(textattrs) # perform cleans
        attr.checkattrs(textattrs, [textattr, trafo.trafo_pt, style.fillstyle])
        trafos = attr.getattrs(textattrs, [trafo.trafo_pt])
        fillstyles = attr.getattrs(textattrs, [style.fillstyle])
        textattrs = attr.getattrs(textattrs, [textattr])
        mathmode = bool(attr.getattrs(textattrs, [_mathmode]))

        if isinstance(text, MultiEngineText):
            text = text.unicode
        output = unicodetextbox_pt(x_pt, y_pt, text, self.font, self.size, mathmode=mathmode)
        for ta in textattrs: # reverse?!
            ta.apply_trafo(output)

        return output

    def text(self, x, y, *args, **kwargs):
        return self.text_pt(unit.topt(x), unit.topt(y), *args, **kwargs)


# from pyx.font.otffile import OpenTypeFont
# 
# class OTFUnicodeText:
# 
#     def __init__(self, fontname, size=10):
#         self.font = OpenTypeFont(config.open(fontname, [config.format.ttf]))
#         self.size = size
# 
#     def preamble(self):
#         raise NotImplemented()
# 
#     def reset(self):
#         raise NotImplemented()
# 
#     def text_pt(self, x_pt, y_pt, text, textattrs=[], texmessages=[], fontmap=None, singlecharmode=False):
#     # def text_pt(self, x_pt, y_pt, text, *args, **kwargs):
#         return unicodetextbox_pt(x_pt, y_pt, text, self.font, self.size)
# 
#     def text(self, x, y, *args, **kwargs):
#         return self.text_pt(unit.topt(x), unit.topt(y), *args, **kwargs)


# old, deprecated names:
texrunner = TexRunner = TexEngine
latexrunner = LatexRunner = LatexEngine

# module level interface documentation for autodoc
# the actual values are setup by the set function

#: the current :class:`MultiEngine` instance for the module level functions
defaulttextengine = None

#: defaulttextengine.preamble (bound method)
preamble = None

#: defaulttextengine.text_pt (bound method)
text_pt = None

#: defaulttextengine.text (bound method)
text = None

#: defaulttextengine.reset (bound method)
reset = None

def set(engine=None, cls=None, mode=None, *args, **kwargs):
    """Setup a new module level :class:`MultiEngine`

    :param engine: the module level engine object to be used, i.e.
        :class:`TexEngine`, :class:`LatexEngine`, or :class:`UnicodeEngine`
    :type cls: Engine object, not instance
    :param cls: identical to *engine*

        :deprecated: use the engine argument instead
    :param mode: ``"tex"`` for :class:`TexEngine` or ``"latex"`` for
        :class:`LatexEngine` with arbitraty capitalization

        :deprecated: use the engine argument instead
    :type mode: str or None
    :param list args: args at class instantiation
    :param dict kwargs: keyword args at at class instantiation

    """
    global defaulttextengine, reset, preamble, text, text_pt
    if mode is not None:
        logger.warning("mode setting is deprecated, use the engine argument instead")
        assert cls is None
        assert engine is None
        engine = {"tex": TexEngine, "latex": LatexEngine}[mode.lower()]
    elif cls is not None:
        logger.warning("cls setting is deprecated, use the engine argument instead")
        assert not engine
        engine = cls
    defaulttextengine = engine(*args, **kwargs)
    preamble = defaulttextengine.preamble
    text_pt = defaulttextengine.text_pt
    text = defaulttextengine.text
    reset = defaulttextengine.reset

# initialize defaulttextengine
set({"TexEngine": TexEngine, "LatexEngine": LatexEngine, "UnicodeEngine": UnicodeEngine}[config.get('text', 'defaulttextengine', 'TexEngine')])


def escapestring(s, replace={" ": "~",
                             "$": r"\$",
                             "&": r"\&",
                             "#": r"\#",
                             "_": r"\_",
                             "%": r"\%",
                             "^": r"\string^",
                             "~": r"\string~",
                             "<": "{$<$}",
                             ">": "{$>$}",
                             "{": r"{$\{$}",
                             "}": r"{$\}$}",
                             "\\": r"{$\setminus$}",
                             "|": r"{$\mid$}"}):
    "Escapes ASCII characters such that they can be typeset by TeX/LaTeX"""
    i = 0
    while i < len(s):
        if not 32 <= ord(s[i]) < 127:
            raise ValueError("escapestring function handles ascii strings only")
        c = s[i]
        try:
            r = replace[c]
        except KeyError:
            i += 1
        else:
            s = s[:i] + r + s[i+1:]
            i += len(r)
    return s



from pyx import *

text.set(text.LatexEngine)
text.preamble(r"\usepackage{times}")

c = canvas.canvas()
c.text(0, 0, r"\LaTeX{} doesn't need to look like \LaTeX{} all the time.")
c.writeEPSfile("font")
c.writePDFfile("font")
c.writeSVGfile("font")



from pyx import *

unit.set(wscale=10)

c = canvas.canvas()

c.stroke(path.line(0, 0, 1, 0) +
         path.line(1, 0, 1, 1) +
         path.line(1, 1, 0, 1) +
         path.line(0, 1, 0, 0))

c.stroke(path.line(2, 0, 3, 0) <<
         path.line(3, 0, 3, 1) <<
         path.line(3, 1, 2, 1) <<
         path.line(2, 1, 2, 0))

p = path.line(4, 0, 5, 0) << path.line(5, 0, 5, 1) << path.line(5, 1, 4, 1)
p.append(path.closepath())
c.stroke(p)

c.writeEPSfile("addjoin")
c.writePDFfile("addjoin")
c.writeSVGfile("addjoin")



from pyx import *

g = graph.graphxy(width=8, x=graph.axis.nestedbar())
g.plot([graph.data.file("minimal.dat", xname="$0, 0", y=2),
        graph.data.file("minimal.dat", xname="$0, 1", y=3)],
       [graph.style.bar()])
g.writeEPSfile("compare")
g.writePDFfile("compare")
g.writeSVGfile("compare")



from random import random
from pyx import *

g = graph.graphxy(width=8, x=graph.axis.bar())
g.plot(graph.data.file("minimal.dat", xname=0, y=2, stack=3,
                       dy="1+random()", dstack="1+random()",
                       context={"random": random}),
       [graph.style.errorbar(),
        graph.style.stackedbarpos("stack"),
        graph.style.bar([color.rgb.green]),
        graph.style.range({"y": "stack"}),
        graph.style.errorbar()])
g.writeEPSfile("errors")
g.writePDFfile("errors")
g.writeSVGfile("errors")



from pyx import *

g = graph.graphxy(width=8, x=graph.axis.split())
g.plot(graph.data.points([((0, 0.1), 0.1),
                          ((0, 0.5), 0.2),
                          ((0, 0.9), 0.3),
                          ((1, 101), 0.7),
                          ((1, 105), 0.8),
                          ((1, 109), 0.9)], x=1, y=2))
g.writeEPSfile("minimal")
g.writePDFfile("minimal")
g.writeSVGfile("minimal")



from pyx import *

c = canvas.canvas()

p = path.path(path.moveto(-2, 0), path.curveto(-1, 0, -1, 1, 0, 1), path.curveto(1, 1, 1, 0, 2, 0))

c.stroke(p, [deco.curvedtext(r"\PyX{} is fun!"),
             deco.curvedtext("left", textattrs=[text.halign.left, text.vshift.mathaxis], arclenfrombegin=0.5, exclude=0.1),
             deco.curvedtext("right", textattrs=[text.halign.right, text.vshift.mathaxis], arclenfromend=0.5, exclude=0.1)])

c.writeEPSfile("textalongpath")
c.writePDFfile("textalongpath")
c.writeSVGfile("textalongpath")



import random
from pyx import *

# introduce a new symbol style where size and color are given by some data
class changesymbol(graph.style.symbol):

    def __init__(self, sizecolumnname="size", colorcolumnname="color",
                       gradient=color.gradient.Rainbow,
                       symbol=graph.style.symbol.circle,
                       symbolattrs=[deco.filled, deco.stroked([color.gray.black])],
                       **kwargs):
        # add some configuration parameters and modify some other
        self.sizecolumnname = sizecolumnname
        self.colorcolumnname = colorcolumnname
        self.gradient = gradient
        graph.style.symbol.__init__(self, symbol=symbol, symbolattrs=symbolattrs, **kwargs)

    def columnnames(self, privatedata, sharedata, agraph, columnnames, dataaxisnames):
        # register the new column names
        if self.sizecolumnname not in columnnames:
            raise ValueError("column '%s' missing" % self.sizecolumnname)
        if self.colorcolumnname not in columnnames:
            raise ValueError("column '%s' missing" % self.colorcolumnname)
        return ([self.sizecolumnname, self.colorcolumnname] +
                graph.style.symbol.columnnames(self, privatedata, sharedata, agraph,
                                               columnnames, dataaxisnames))

    def drawpoint(self, privatedata, sharedata, graph, point):
        # replace the original drawpoint method by a slightly revised one
        if sharedata.vposvalid and privatedata.symbolattrs is not None:
            x_pt, y_pt = graph.vpos_pt(*sharedata.vpos)
            color = self.gradient.getcolor(point[self.colorcolumnname])
            privatedata.symbol(privatedata.symbolcanvas, x_pt, y_pt,
                               privatedata.size_pt*point[self.sizecolumnname],
                               privatedata.symbolattrs + [color])

g = graph.graphxy(width=10)
g.plot(graph.data.points([[random.random() for i in range(4)]
                          for i in range(1000)],
                         x=1, y=2, size=3, color=4),
       [changesymbol()])
g.writeEPSfile("changesymbol")
g.writePDFfile("changesymbol")
g.writeSVGfile("changesymbol")



import fractions, math
from pyx import *
from pyx.graph import axis

# we here use parters and texters which are explained in the examples below
log2parter = axis.parter.log([axis.parter.preexp([axis.tick.rational(1)], 4),
                              axis.parter.preexp([axis.tick.rational(1)], 2)])
log2texter = axis.texter.default(base=fractions.Fraction(2))

g = graph.graphxy(width=10,
    x=axis.log(min=1, max=1024),
    y=axis.log(min=1, max=1024, parter=log2parter),
    y2=axis.log(min=1, max=1024, parter=log2parter, texter=log2texter))

g.writeEPSfile("log")
g.writePDFfile("log")
g.writeSVGfile("log")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2007-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


class metric:

    def width_ds(self, glyphname):
        raise NotImplementedError()

    def width_pt(self, glyphnames, size_pt):
        raise NotImplementedError()

    def height_pt(self, glyphnames, size_pt):
        raise NotImplementedError()

    def depth_pt(self, glyphnames, size_pt):
        raise NotImplementedError()

    def resolveligatures(self, glyphnames):
        return glyphnames

    def resolvekernings(self, glyphnames, size_pt=None):
        result = [None]*(2*len(glyphnames)-1)
        for i, glyphname in enumerate(glyphnames):
            result[2*i] = glyphname
        return result



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2007-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


from pyx import reader

class ENFfileError(Exception):
    pass

class ENCfile:

    def __init__(self, bytes):
        c = reader.PStokenizer(bytes, "")

        # name of encoding
        self.name = c.gettoken()
        token = c.gettoken()
        if token != "[":
            raise ENCfileError("cannot parse encoding file '%s', expecting '[' got '%s'" % (filename, token))
        self.vector = []
        for i in range(256):
            token = c.gettoken()
            if token == "]":
                raise ENCfileError("not enough charcodes in encoding file '%s'" % filename)
            if not token[0] == "/":
                raise ENCfileError("token does not start with / in encoding file '%s'" % filename)
            self.vector.append(token[1:])
        if c.gettoken() != "]":
            raise ENCfileError("too many charcodes in encoding file '%s'" % filename)
        token = c.gettoken()
        if token != "def":
            raise ENCfileError("cannot parse encoding file '%s', expecting 'def' got '%s'" % (filename, token))




# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2007 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2002-2007 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from . import attr

class canvasitem:

    """Base class for everything which can be inserted into a canvas"""

    def bbox(self):
        """return bounding box of canvasitem"""
        raise NotImplementedError()

    def requiretextregion(self):
        """indicates whether a canvasitem needs to be part of a PDF text
        region"""
        return False

    def processPS(self, file, writer, context, registry, bbox):
        """process canvasitem by writing the corresponding PS code to file and
        by updating context, registry as well as bbox

        - the PS code corresponding to the canvasitem has to be written in the
          stream file, which provides a write(string) method
        - writer is the PSwriter used for the output
        - context is an instance of pswriter.context which is used for keeping
          track of the graphics state (current linewidth, colorspace and font,
          etc.)
        - registry is used for tracking resources needed by the canvasitem
        - bbox has to be updated to include the bounding box of the canvasitem
        """
        raise NotImplementedError()

    def processPDF(self, file, writer, context, registry, bbox):
        """process canvasitem by writing the corresponding PDF code to file and
        by updating context, registry as well as bbox

        - the PDF code corresponding to the canvasitem has to be written in the
          stream file, which provides a write(string) method
        - writer is the PDFwriter used for the output, which contains properties
          like whether streamcompression is used
        - context is an instance of pdfwriter.context which is used for keeping
          track of the graphics state, in particular for the emulation of PS
          behaviour regarding fill and stroke styles, for keeping track of the
          currently selected font as well as of text regions.
        - registry is used for tracking resources needed by the canvasitem
        - bbox has to be updated to include the bounding box of the canvasitem
        """
        raise NotImplementedError()


class deformer(attr.attr):

    def deform(self, basepath):
        raise NotImplementedError()




#!/usr/bin/env python
from pyx import *

g = graph.graphxyz(size=4, projector=graph.graphxyz.parallel(170, 45))
g.plot(graph.data.file("color.dat", x=1, y=2, z=3, color=3),
       [graph.style.surface(gradient=color.gradient.RedGreen,
                            gridcolor=color.rgb.black,
                            backcolor=color.rgb.black)])
g.writeEPSfile("color")
g.writePDFfile("color")
g.writeSVGfile("color")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2005-2011 André Wobst <wobsta@pyx-project.org>
# Copyright (C) 2006-2011 Jörg Lehmann <joerg@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import array, binascii, io, logging, math, re
try:
    import zlib
    haszlib = True
except ImportError:
    haszlib = False

logger = logging.getLogger("pyx")

from pyx import trafo, reader, writer
from pyx.path import path, moveto_pt, lineto_pt, curveto_pt, closepath

try:
    from ._t1code import *
except Exception:
    from .t1code import *


adobestandardencoding = [None, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None,
                         "space", "exclam", "quotedbl", "numbersign", "dollar", "percent", "ampersand", "quoteright",
                         "parenleft", "parenright", "asterisk", "plus", "comma", "hyphen", "period", "slash",
                         "zero", "one", "two", "three", "four", "five", "six", "seven",
                         "eight", "nine", "colon", "semicolon", "less", "equal", "greater", "question",
                         "at", "A", "B", "C", "D", "E", "F", "G",
                         "H", "I", "J", "K", "L", "M", "N", "O",
                         "P", "Q", "R", "S", "T", "U", "V", "W",
                         "X", "Y", "Z", "bracketleft", "backslash", "bracketright", "asciicircum", "underscore",
                         "quoteleft", "a", "b", "c", "d", "e", "f", "g",
                         "h", "i", "j", "k", "l", "m", "n", "o",
                         "p", "q", "r", "s", "t", "u", "v", "w",
                         "x", "y", "z", "braceleft", "bar", "braceright", "asciitilde", None,
                         None, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None,
                         None, "exclamdown", "cent", "sterling", "fraction", "yen", "florin", "section",
                         "currency", "quotesingle", "quotedblleft", "guillemotleft", "guilsinglleft", "guilsinglright", "fi", "fl",
                         None, "endash", "dagger", "daggerdbl", "periodcentered", None, "paragraph", "bullet",
                         "quotesinglbase", "quotedblbase", "quotedblright", "guillemotright", "ellipsis", "perthousand", None, "questiondown",
                         None, "grave", "acute", "circumflex", "tilde", "macron", "breve", "dotaccent",
                         "dieresis", None, "ring", "cedilla", None, "hungarumlaut", "ogonek", "caron",
                         "emdash", None, None, None, None, None, None, None,
                         None, None, None, None, None, None, None, None,
                         None, "AE", None, "ordfeminine", None, None, None, None,
                         "Lslash", "Oslash", "OE", "ordmasculine", None, None, None, None,
                         None, "ae", None, None, None, "dotlessi", None, None,
                         "lslash", "oslash", "oe", "germandbls", None, None, None, None]

class T1context:

    def __init__(self, t1font, flex=True):
        """context for T1cmd evaluation"""
        self.t1font = t1font

        # state description
        self.x = None
        self.y = None
        self.wx = None
        self.wy = None
        self.t1stack = []
        self.psstack = []
        self.flex = flex


######################################################################
# T1 commands
# Note, that the T1 commands are variable-free except for plain number,
# which are stored as integers. All other T1 commands exist as a single
# instance only

T1cmds = {}
T1subcmds = {}

class T1cmd:

    def __init__(self, code, subcmd=0):
        self.code = code
        self.subcmd = subcmd
        if subcmd:
            T1subcmds[code] = self
        else:
            T1cmds[code] = self

    def __str__(self):
        """returns a string representation of the T1 command"""
        raise NotImplementedError

    def updatepath(self, path, trafo, context):
        """update path instance applying trafo to the points"""
        raise NotImplementedError

    def gathercalls(self, seacglyphs, subrs, context):
        """gather dependancy information

        subrs is the "called-subrs" dictionary. gathercalls will insert the
        subr number as key having the value 1, i.e. subrs will become the
        numbers of used subrs. Similar seacglyphs will contain all glyphs in
        composite characters (subrs for those glyphs will also
        already be included).

        This method might will not properly update all information in the
        context (especially consuming values from the stack) and will also skip
        various tests for performance reasons. For most T1 commands it just
        doesn't need to do anything.
        """
        pass


# commands for starting and finishing

class _T1endchar(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 14)

    def __str__(self):
        return "endchar"

    def updatepath(self, path, trafo, context):
        pass

T1endchar = _T1endchar()


class _T1hsbw(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 13)

    def __str__(self):
        return "hsbw"

    def updatepath(self, path, trafo, context):
        sbx = context.t1stack.pop(0)
        wx = context.t1stack.pop(0)
        path.append(moveto_pt(*trafo.apply_pt(sbx, 0)))
        context.x = sbx
        context.y = 0
        context.wx = wx
        context.wy = 0

T1hsbw = _T1hsbw()


class _T1seac(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 6, subcmd=1)

    def __str__(self):
        return "seac"

    def updatepath(self, path, atrafo, context):
        sab = context.t1stack.pop(0)
        adx = context.t1stack.pop(0)
        ady = context.t1stack.pop(0)
        bchar = context.t1stack.pop(0)
        achar = context.t1stack.pop(0)
        aglyph = adobestandardencoding[achar]
        bglyph = adobestandardencoding[bchar]
        context.t1font.updateglyphpath(bglyph, path, atrafo, context)
        atrafo = atrafo * trafo.translate_pt(adx-sab, ady)
        context.t1font.updateglyphpath(aglyph, path, atrafo, context)

    def gathercalls(self, seacglyphs, subrs, context):
        achar = context.t1stack.pop()
        bchar = context.t1stack.pop()
        aglyph = adobestandardencoding[achar]
        bglyph = adobestandardencoding[bchar]
        seacglyphs.add(aglyph)
        seacglyphs.add(bglyph)
        context.t1font.gatherglyphcalls(bglyph, seacglyphs, subrs, context)
        context.t1font.gatherglyphcalls(aglyph, seacglyphs, subrs, context)

T1seac = _T1seac()


class _T1sbw(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 7, subcmd=1)

    def __str__(self):
        return "sbw"

    def updatepath(self, path, trafo, context):
        sbx = context.t1stack.pop(0)
        sby = context.t1stack.pop(0)
        wx = context.t1stack.pop(0)
        wy = context.t1stack.pop(0)
        path.append(moveto_pt(*trafo.apply_pt(sbx, sby)))
        context.x = sbx
        context.y = sby
        context.wx = wx
        context.wy = wy

T1sbw = _T1sbw()


# path construction commands

class _T1closepath(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 9)

    def __str__(self):
        return "closepath"

    def updatepath(self, path, trafo, context):
        path.append(closepath())
        # The closepath in T1 is different from PostScripts in that it does
        # *not* modify the current position; hence we need to add an additional
        # moveto here ...
        path.append(moveto_pt(*trafo.apply_pt(context.x, context.y)))

T1closepath = _T1closepath()


class _T1hlineto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 6)

    def __str__(self):
        return "hlineto"

    def updatepath(self, path, trafo, context):
        dx = context.t1stack.pop(0)
        path.append(lineto_pt(*trafo.apply_pt(context.x + dx, context.y)))
        context.x += dx

T1hlineto = _T1hlineto()


class _T1hmoveto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 22)

    def __str__(self):
        return "hmoveto"

    def updatepath(self, path, trafo, context):
        dx = context.t1stack.pop(0)
        path.append(moveto_pt(*trafo.apply_pt(context.x + dx, context.y)))
        context.x += dx

T1hmoveto = _T1hmoveto()


class _T1hvcurveto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 31)

    def __str__(self):
        return "hvcurveto"

    def updatepath(self, path, trafo, context):
        dx1 = context.t1stack.pop(0)
        dx2 = context.t1stack.pop(0)
        dy2 = context.t1stack.pop(0)
        dy3 = context.t1stack.pop(0)
        path.append(curveto_pt(*(trafo.apply_pt(context.x + dx1,       context.y) +
                                 trafo.apply_pt(context.x + dx1 + dx2, context.y + dy2) +
                                 trafo.apply_pt(context.x + dx1 + dx2, context.y + dy2 + dy3))))
        context.x += dx1+dx2
        context.y += dy2+dy3

T1hvcurveto = _T1hvcurveto()


class _T1rlineto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 5)

    def __str__(self):
        return "rlineto"

    def updatepath(self, path, trafo, context):
        dx = context.t1stack.pop(0)
        dy = context.t1stack.pop(0)
        path.append(lineto_pt(*trafo.apply_pt(context.x + dx, context.y + dy)))
        context.x += dx
        context.y += dy

T1rlineto = _T1rlineto()


class _T1rmoveto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 21)

    def __str__(self):
        return "rmoveto"

    def updatepath(self, path, trafo, context):
        dx = context.t1stack.pop(0)
        dy = context.t1stack.pop(0)
        path.append(moveto_pt(*trafo.apply_pt(context.x + dx, context.y + dy)))
        context.x += dx
        context.y += dy

T1rmoveto = _T1rmoveto()


class _T1rrcurveto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 8)

    def __str__(self):
        return "rrcurveto"

    def updatepath(self, path, trafo, context):
        dx1 = context.t1stack.pop(0)
        dy1 = context.t1stack.pop(0)
        dx2 = context.t1stack.pop(0)
        dy2 = context.t1stack.pop(0)
        dx3 = context.t1stack.pop(0)
        dy3 = context.t1stack.pop(0)
        path.append(curveto_pt(*(trafo.apply_pt(context.x + dx1,             context.y + dy1) +
                                 trafo.apply_pt(context.x + dx1 + dx2,       context.y + dy1 + dy2) +
                                 trafo.apply_pt(context.x + dx1 + dx2 + dx3, context.y + dy1 + dy2 + dy3))))
        context.x += dx1+dx2+dx3
        context.y += dy1+dy2+dy3

T1rrcurveto = _T1rrcurveto()


class _T1vlineto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 7)

    def __str__(self):
        return "vlineto"

    def updatepath(self, path, trafo, context):
        dy = context.t1stack.pop(0)
        path.append(lineto_pt(*trafo.apply_pt(context.x, context.y + dy)))
        context.y += dy

T1vlineto = _T1vlineto()


class _T1vmoveto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 4)

    def __str__(self):
        return "vmoveto"

    def updatepath(self, path, trafo, context):
        dy = context.t1stack.pop(0)
        path.append(moveto_pt(*trafo.apply_pt(context.x, context.y + dy)))
        context.y += dy

T1vmoveto = _T1vmoveto()


class _T1vhcurveto(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 30)

    def __str__(self):
        return "vhcurveto"

    def updatepath(self, path, trafo, context):
        dy1 = context.t1stack.pop(0)
        dx2 = context.t1stack.pop(0)
        dy2 = context.t1stack.pop(0)
        dx3 = context.t1stack.pop(0)
        path.append(curveto_pt(*(trafo.apply_pt(context.x,             context.y + dy1) +
                                 trafo.apply_pt(context.x + dx2,       context.y + dy1 + dy2) +
                                 trafo.apply_pt(context.x + dx2 + dx3, context.y + dy1 + dy2))))
        context.x += dx2+dx3
        context.y += dy1+dy2

T1vhcurveto = _T1vhcurveto()


# hint commands

class _T1dotsection(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 0, subcmd=1)

    def __str__(self):
        return "dotsection"

    def updatepath(self, path, trafo, context):
        pass

T1dotsection = _T1dotsection()


class _T1hstem(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 1)

    def __str__(self):
        return "hstem"

    def updatepath(self, path, trafo, context):
        y = context.t1stack.pop(0)
        dy = context.t1stack.pop(0)

T1hstem = _T1hstem()


class _T1hstem3(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 2, subcmd=1)

    def __str__(self):
        return "hstem3"

    def updatepath(self, path, trafo, context):
        y0 = context.t1stack.pop(0)
        dy0 = context.t1stack.pop(0)
        y1 = context.t1stack.pop(0)
        dy1 = context.t1stack.pop(0)
        y2 = context.t1stack.pop(0)
        dy2 = context.t1stack.pop(0)

T1hstem3 = _T1hstem3()


class _T1vstem(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 3)

    def __str__(self):
        return "vstem"

    def updatepath(self, path, trafo, context):
        x = context.t1stack.pop(0)
        dx = context.t1stack.pop(0)

T1vstem = _T1vstem()


class _T1vstem3(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 1, subcmd=1)

    def __str__(self):
        return "vstem3"

    def updatepath(self, path, trafo, context):
        self.x0 = context.t1stack.pop(0)
        self.dx0 = context.t1stack.pop(0)
        self.x1 = context.t1stack.pop(0)
        self.dx1 = context.t1stack.pop(0)
        self.x2 = context.t1stack.pop(0)
        self.dx2 = context.t1stack.pop(0)

T1vstem3 = _T1vstem3()


# arithmetic command

class _T1div(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 12, subcmd=1)

    def __str__(self):
        return "div"

    def updatepath(self, path, trafo, context):
        num2 = context.t1stack.pop()
        num1 = context.t1stack.pop()
        context.t1stack.append(divmod(num1, num2)[0])

    def gathercalls(self, seacglyphs, subrs, context):
        num2 = context.t1stack.pop()
        num1 = context.t1stack.pop()
        context.t1stack.append(divmod(num1, num2)[0])

T1div = _T1div()


# subroutine commands

class _T1callothersubr(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 16, subcmd=1)

    def __str__(self):
        return "callothersubr"

    def updatepath(self, path, trafo, context):
        othersubrnumber = context.t1stack.pop()
        n = context.t1stack.pop()
        for i in range(n):
            context.psstack.append(context.t1stack.pop(0))
        if othersubrnumber == 0:
            flex_size, x, y = context.psstack[-3:]
            if context.flex:
                x1, y1, x2, y2, x3, y3 = context.psstack[2:8]
                x1, y1 = trafo.apply_pt(x1, y1)
                x2, y2 = trafo.apply_pt(x2, y2)
                x3, y3 = trafo.apply_pt(x3, y3)
                path.append(curveto_pt(x1, y1, x2, y2, x3, y3))
                x1, y1, x2, y2, x3, y3 = context.psstack[8:14]
                x1, y1 = trafo.apply_pt(x1, y1)
                x2, y2 = trafo.apply_pt(x2, y2)
                x3, y3 = trafo.apply_pt(x3, y3)
                path.append(curveto_pt(x1, y1, x2, y2, x3, y3))
            else:
                path.append(lineto_pt(*trafo.apply_pt(x, y)))
            context.psstack = [y, x]
        elif othersubrnumber == 1:
            pass
        elif othersubrnumber == 2:
            path.pathitems.pop()
            context.psstack.append(context.x)
            context.psstack.append(context.y)

    def gathercalls(self, seacglyphs, subrs, context):
        othersubrnumber = context.t1stack.pop()
        n = context.t1stack.pop()
        context.psstack.extend([context.t1stack.pop() for i in range(n)][::-1])

T1callothersubr = _T1callothersubr()


class _T1callsubr(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 10)

    def __str__(self):
        return "callsubr"

    def updatepath(self, path, trafo, context):
        subr = context.t1stack.pop()
        context.t1font.updatesubrpath(subr, path, trafo, context)

    def gathercalls(self, seacglyphs, subrs, context):
        subr = context.t1stack.pop()
        subrs.add(subr)
        context.t1font.gathersubrcalls(subr, seacglyphs, subrs, context)

T1callsubr = _T1callsubr()


class _T1pop(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 17, subcmd=1)

    def __str__(self):
        return "pop"

    def updatepath(self, path, trafo, context):
        context.t1stack.append(context.psstack.pop())

    def gathercalls(self, seacglyphs, subrs, context):
        context.t1stack.append(context.psstack.pop())

T1pop = _T1pop()


class _T1return(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 11)

    def __str__(self):
        return "return"

    def updatepath(self, path, trafo, context):
        pass

T1return = _T1return()


class _T1setcurrentpoint(T1cmd):

    def __init__(self):
        T1cmd.__init__(self, 33, subcmd=1)

    def __str__(self):
        return "setcurrentpoint"

    def updatepath(self, path, trafo, context):
        context.x = context.t1stack.pop(0)
        context.y = context.t1stack.pop(0)

T1setcurrentpoint = _T1setcurrentpoint()


######################################################################

class FontFormatError(Exception):
    pass

class T1File:

    eexecr = 55665
    charstringr = 4330

    fontnamepattern = re.compile(r"/FontName\s+/(.*?)\s+def\s+")
    fontmatrixpattern = re.compile(r"/FontMatrix\s*\[\s*(-?[0-9.]+)\s+(-?[0-9.]+)\s+(-?[0-9.]+)\s+(-?[0-9.]+)\s+(-?[0-9.]+)\s+(-?[0-9.]+)\s*\]\s*(readonly\s+)?def")

    def __init__(self, data1, data2eexec, data3):
        """initializes a t1font instance

        data1 and data3 are the two clear text data parts and data2 is
        the binary data part"""
        self.data1 = data1
        self._data2eexec = data2eexec
        self.data3 = data3

        # marker and value for decoded data
        self._data2 = None
        # note that data2eexec is set to none by setsubrcmds and setglyphcmds
        # this *also* denotes, that data2 is out-of-date; hence they are both
        # marked by an _ and getdata2 and getdata2eexec will properly resolve
        # the current state of decoding ...

        # marker and value for standard encoding check
        self.encoding = None

        self.name, = self.fontnamepattern.search(self.data1).groups()
        m11, m12, m21, m22, v1, v2 = list(map(float, self.fontmatrixpattern.search(self.data1).groups()[:6]))
        self.fontmatrix = trafo.trafo_pt(matrix=((m11, m12), (m21, m22)), vector=(v1, v2))

    def _eexecdecode(self, code):
        """eexec decoding of code"""
        return decoder(code, self.eexecr, 4)

    def _charstringdecode(self, code):
        """charstring decoding of code"""
        return decoder(code, self.charstringr, self.lenIV)

    def _eexecencode(self, data):
        """eexec encoding of data"""
        return encoder(data, self.eexecr, b"PyX!")

    def _charstringencode(self, data):
        """eexec encoding of data"""
        return encoder(data, self.charstringr, b"PyX!"[:self.lenIV])

    def _encoding(self):
        """helper method to lookup the encoding in the font"""
        c = reader.PStokenizer(self.data1, "/Encoding")
        token1 = c.gettoken()
        token2 = c.gettoken()
        if token1 == "StandardEncoding" and token2 == "def":
            self.encoding = adobestandardencoding
        else:
            self.encoding = [None]*256
            while True:
                self.encodingstart = c.pos
                if c.gettoken() == "dup":
                    break
            while True:
                i = c.getint()
                glyph = c.gettoken()
                if 0 <= i < 256:
                    self.encoding[i] = glyph[1:]
                token = c.gettoken(); assert token == "put"
                self.encodingend = c.pos
                token = c.gettoken()
                if token == "readonly" or token == "def":
                    break
                assert token == "dup"

    lenIVpattern = re.compile(rb"/lenIV\s+(\d+)\s+def\s+")
    flexhintsubrs = [[3, 0, T1callothersubr, T1pop, T1pop, T1setcurrentpoint, T1return],
                     [0, 1, T1callothersubr, T1return],
                     [0, 2, T1callothersubr, T1return],
                     [T1return]]

    def _data2decode(self):
        """decodes data2eexec to the data2 string and the subr and glyphs dictionary

        It doesn't make sense to call this method twice -- check the content of
        data2 before calling. The method also keeps the subrs and charstrings
        start and end positions for later use."""
        self._data2 = self._eexecdecode(self._data2eexec)

        m = self.lenIVpattern.search(self._data2)
        if m:
            self.lenIV = int(m.group(1))
        else:
            self.lenIV = 4

        self.emptysubr = self._charstringencode(b"\x0b") # 11, i.e. return

        # extract Subrs
        c = reader.PSbytes_tokenizer(self._data2, b"/Subrs")
        self.subrsstart = c.pos
        arraycount = c.getint()
        token = c.gettoken(); assert token == b"array"
        self.subrs = []
        for i in range(arraycount):
            token = c.gettoken(); assert token == b"dup"
            token = c.getint(); assert token == i
            size = c.getint()
            if not i:
                self.subrrdtoken = c.gettoken()
            else:
                token = c.gettoken(); assert token == self.subrrdtoken
            self.subrs.append(c.getbytes(size))
            token = c.gettoken()
            if token == b"noaccess":
                token = token + b" " + c.gettoken()
            if not i:
                self.subrnptoken = token
            else:
                assert token == self.subrnptoken
        self.subrsend = c.pos

        # hasflexhintsubrs is a boolean indicating that the font uses flex or
        # hint replacement subrs as specified by Adobe (tm). When it does, the
        # first 4 subrs should all be copied except when none of them are used
        # in the stripped version of the font since we then get a font not
        # using flex or hint replacement subrs at all.
        self.hasflexhintsubrs = (arraycount >= len(self.flexhintsubrs) and
                                 [self.getsubrcmds(i)
                                  for i in range(len(self.flexhintsubrs))] == self.flexhintsubrs)

        # extract glyphs
        self.glyphs = {}
        self.glyphlist = [] # we want to keep the order of the glyph names
        c = reader.PSbytes_tokenizer(self._data2, b"/CharStrings")
        self.charstringsstart = c.pos
        c.getint()
        token = c.gettoken(); assert token == b"dict"
        token = c.gettoken(); assert token == b"dup"
        token = c.gettoken(); assert token == b"begin"
        first = True
        while True:
            chartoken = c.gettoken().decode("ascii")
            if chartoken == "end":
                break
            assert chartoken[0] == "/"
            size = c.getint()
            if first:
                self.glyphrdtoken = c.gettoken()
            else:
                token = c.gettoken(); assert token == self.glyphrdtoken
            self.glyphlist.append(chartoken[1:])
            self.glyphs[chartoken[1:]] = c.getbytes(size)
            if first:
                self.glyphndtoken = c.gettoken()
            else:
                token = c.gettoken(); assert token == self.glyphndtoken
            first = False
        self.charstringsend = c.pos
        assert not self.subrs or self.subrrdtoken == self.glyphrdtoken

    def _cmds(self, code):
        """return a list of T1cmd's for encoded charstring data in code"""
        code = array.array("B", self._charstringdecode(code))
        cmds = []
        while code:
            x = code.pop(0)
            if x == 12: # this starts an escaped cmd
                cmds.append(T1subcmds[code.pop(0)])
            elif 0 <= x < 32: # those are cmd's
                cmds.append(T1cmds[x])
            elif 32 <= x <= 246: # short ints
                cmds.append(x-139)
            elif 247 <= x <= 250: # mid size ints
                cmds.append(((x - 247)*256) + code.pop(0) + 108)
            elif 251 <= x <= 254: # mid size ints
                cmds.append(-((x - 251)*256) - code.pop(0) - 108)
            else: # x = 255, i.e. full size ints
                y = ((code.pop(0)*256+code.pop(0))*256+code.pop(0))*256+code.pop(0)
                if y > (1 << 31):
                    cmds.append(y - (1 << 32))
                else:
                    cmds.append(y)
        return cmds

    def _code(self, cmds):
        """return an encoded charstring data for list of T1cmd's in cmds"""
        code = array.array("B")
        for cmd in cmds:
            try:
                if cmd.subcmd:
                    code.append(12)
                code.append(cmd.code)
            except AttributeError:
                if -107 <= cmd <= 107:
                    code.append(cmd+139)
                elif 108 <= cmd <= 1131:
                    a, b = divmod(cmd-108, 256)
                    code.append(a+247)
                    code.append(b)
                elif -1131 <= cmd <= -108:
                    a, b = divmod(-cmd-108, 256)
                    code.append(a+251)
                    code.append(b)
                else:
                    if cmd < 0:
                        cmd += 1 << 32
                    cmd, x4 = divmod(cmd, 256)
                    cmd, x3 = divmod(cmd, 256)
                    x1, x2 = divmod(cmd, 256)
                    code.append(255)
                    code.append(x1)
                    code.append(x2)
                    code.append(x3)
                    code.append(x4)
        return self._charstringencode(code.tobytes())

    def getsubrcmds(self, subr):
        """return a list of T1cmd's for subr subr"""
        if not self._data2:
            self._data2decode()
        return self._cmds(self.subrs[subr])

    def getglyphcmds(self, glyph):
        """return a list of T1cmd's for glyph glyph"""
        if not self._data2:
            self._data2decode()
        return self._cmds(self.glyphs[glyph])

    def setsubrcmds(self, subr, cmds):
        """replaces the T1cmd's by the list cmds for subr subr"""
        if not self._data2:
            self._data2decode()
        self._data2eexec = None
        self.subrs[subr] = self._code(cmds)

    def setglyphcmds(self, glyph, cmds):
        """replaces the T1cmd's by the list cmds for glyph glyph"""
        if not self._data2:
            self._data2decode()
        self._data2eexec = None
        self.glyphs[glyph] = self._code(cmds)

    def updatepath(self, cmds, path, trafo, context):
        for cmd in cmds:
            if isinstance(cmd, T1cmd):
                cmd.updatepath(path, trafo, context)
            else:
                context.t1stack.append(cmd)

    def updatesubrpath(self, subr, path, trafo, context):
        self.updatepath(self.getsubrcmds(subr), path, trafo, context)

    def updateglyphpath(self, glyph, path, trafo, context):
        self.updatepath(self.getglyphcmds(glyph), path, trafo, context)

    def gathercalls(self, cmds, seacglyphs, subrs, context):
        for cmd in cmds:
            if isinstance(cmd, T1cmd):
                cmd.gathercalls(seacglyphs, subrs, context)
            else:
                context.t1stack.append(cmd)

    def gathersubrcalls(self, subr, seacglyphs, subrs, context):
        self.gathercalls(self.getsubrcmds(subr), seacglyphs, subrs, context)

    def gatherglyphcalls(self, glyph, seacglyphs, subrs, context):
        self.gathercalls(self.getglyphcmds(glyph), seacglyphs, subrs, context)

    def getglyphpath_pt(self, x_pt, y_pt, glyph, size_pt, convertcharcode=False, flex=True):
        """return an object containing the PyX path, wx_pt and wy_pt for glyph named glyph"""
        if convertcharcode:
            if not self.encoding:
                self._encoding()
            glyph = self.encoding[glyph]
        t = self.fontmatrix.scaled(size_pt)
        tpath = t.translated_pt(x_pt, y_pt)
        context = T1context(self, flex=flex)
        p = path()
        self.updateglyphpath(glyph, p, tpath, context)
        class glyphpath:
            def __init__(self, p, wx_pt, wy_pt):
                self.path = p
                self.wx_pt = wx_pt
                self.wy_pt = wy_pt
        return glyphpath(p, *t.apply_pt(context.wx, context.wy))

    def getdata2(self, subrs=None, glyphs=None):
        """makes a data2 string

        subrs is a dict containing those subrs numbers as keys,
        which are to be contained in the subrsstring to be created.
        If subrs is None, all subrs in self.subrs will be used.
        The subrs dict might be modified *in place*.

        glyphs is a dict containing those glyph names as keys,
        which are to be contained in the charstringsstring to be created.
        If glyphs is None, all glyphs in self.glyphs will be used."""
        w = writer.writer(io.BytesIO())

        def addsubrs(subrs):
            if subrs is not None:
                # some adjustments to the subrs dict
                if subrs:
                    subrsmin = min(subrs)
                    subrsmax = max(subrs)
                    if self.hasflexhintsubrs and subrsmin < len(self.flexhintsubrs):
                        # According to the spec we need to keep all the flex and hint subrs
                        # as long as any of it is used.
                        for subr in range(len(self.flexhintsubrs)):
                            subrs.add(subr)
                        subrsmax = max(subrs)
                else:
                    subrsmax = -1
            else:
                # build a new subrs dict containing all subrs
                subrs = dict([(subr, 1) for subr in range(len(self.subrs))])
                subrsmax = len(self.subrs) - 1

            # build the string from all selected subrs
            w.write("%d array\n" % (subrsmax + 1))
            for subr in range(subrsmax+1):
                if subr in subrs:
                    code = self.subrs[subr]
                else:
                    code = self.emptysubr
                w.write("dup %d %d " % (subr, len(code)))
                w.write_bytes(self.subrrdtoken)
                w.write_bytes(b" ")
                w.write_bytes(code)
                w.write_bytes(b" ")
                w.write_bytes(self.subrnptoken)
                w.write_bytes(b"\n")

        def addcharstrings(glyphs):
            w.write("%d dict dup begin\n" % (glyphs is None and len(self.glyphlist) or len(glyphs)))
            for glyph in self.glyphlist:
                if glyphs is None or glyph in glyphs:
                    w.write("/%s %d " % (glyph, len(self.glyphs[glyph])))
                    w.write_bytes(self.glyphrdtoken)
                    w.write_bytes(b" ")
                    w.write_bytes(self.glyphs[glyph])
                    w.write_bytes(b" ")
                    w.write_bytes(self.glyphndtoken)
                    w.write_bytes(b"\n")
            w.write("end\n")

        if self.subrsstart < self.charstringsstart:
            w.write_bytes(self._data2[:self.subrsstart])
            addsubrs(subrs)
            w.write_bytes(self._data2[self.subrsend:self.charstringsstart])
            addcharstrings(glyphs)
            w.write_bytes(self._data2[self.charstringsend:])
        else:
            w.write_bytes(self._data2[:self.charstringsstart])
            addcharstrings(glyphs)
            w.write_bytes(self._data2[self.charstringsend:self.subrsstart])
            addsubrs(subrs)
            w.write_bytes(self._data2[self.subrsend:])
        return w.file.getvalue()

    def getdata2eexec(self):
        if self._data2eexec:
            return self._data2eexec
        # note that self._data2 is out-of-date here too, hence we need to call getdata2
        return self._eexecencode(self.getdata2())

    newlinepattern = re.compile("\\s*[\r\n]\\s*")
    uniqueidstrpattern = re.compile(r"%?/UniqueID\s+\d+\s+def\s+")
    uniqueidbytespattern = re.compile(rb"%?/UniqueID\s+\d+\s+def\s+")
        # when UniqueID is commented out (as in modern latin), prepare to remove the comment character as well

    def getstrippedfont(self, glyphs, charcodes):
        """create a T1File instance containing only certain glyphs

        glyphs is a set of the glyph names. It might be modified *in place*!
        """
        if not self.encoding:
            self._encoding()
        for charcode in charcodes:
            glyphs.add(self.encoding[charcode])

        # collect information about used glyphs and subrs
        seacglyphs = set()
        subrs = set()
        for glyph in glyphs:
            self.gatherglyphcalls(glyph, seacglyphs, subrs, T1context(self))
        # while we have gathered all subrs for the seacglyphs alreadys, we
        # might have missed the glyphs themself (when they are not used stand-alone)
        glyphs.update(seacglyphs)
        glyphs.add(".notdef")

        # strip data1
        if self.encoding is adobestandardencoding:
            data1 = self.data1
        else:
            encodingstrings = []
            for char, glyph in enumerate(self.encoding):
                if glyph in glyphs:
                    encodingstrings.append("dup %i /%s put\n" % (char, glyph))
            data1 = self.data1[:self.encodingstart] + "\n" + "".join(encodingstrings) + self.data1[self.encodingend:]
        data1 = self.newlinepattern.subn("\n", data1)[0]
        data1 = self.uniqueidstrpattern.subn("", data1)[0]

        # strip data2
        data2 = self.uniqueidbytespattern.subn(b"", self.getdata2(subrs, glyphs))[0]

        # strip data3
        data3 = self.newlinepattern.subn("\n", self.data3)[0]

        # create and return the new font instance
        return T1File(data1.rstrip() + "\n", self._eexecencode(data2), data3.rstrip() + "\n")

    # The following two methods, writePDFfontinfo and getglyphinfo,
    # extract informtion which should better be taken from the afm file.
    def writePDFfontinfo(self, file):
        try:
            glyphinfo_y = self.getglyphinfo("y")
            glyphinfo_W = self.getglyphinfo("W")
            glyphinfo_H = self.getglyphinfo("H")
            glyphinfo_h = self.getglyphinfo("h")
            glyphinfo_period = self.getglyphinfo("period")
            glyphinfo_colon = self.getglyphinfo("colon")
        except Exception:
            logger.warning("Auto-guessing of font information for font '%s' failed. We're writing stub data instead." % self.name)
            file.write("/Flags 4\n")
            file.write("/FontBBox [0 -100 1000 1000]\n")
            file.write("/ItalicAngle 0\n")
            file.write("/Ascent 1000\n")
            file.write("/Descent -100\n")
            file.write("/CapHeight 700\n")
            file.write("/StemV 100\n")
        else:
            if not self.encoding:
                self._encoding()
            # As a simple heuristics we assume non-symbolic fonts if and only
            # if the Adobe standard encoding is used. All other font flags are
            # not specified here.
            if self.encoding is adobestandardencoding:
                file.write("/Flags 32\n")
            else:
                file.write("/Flags 4\n")
            file.write("/FontBBox [0 %d %d %d]\n" % (glyphinfo_y[3], glyphinfo_W[0], glyphinfo_H[5]))
            file.write("/ItalicAngle %d\n" % math.degrees(math.atan2(glyphinfo_period[4]-glyphinfo_colon[4], glyphinfo_colon[5]-glyphinfo_period[5])))
            file.write("/Ascent %d\n" % glyphinfo_H[5])
            file.write("/Descent %d\n" % glyphinfo_y[3])
            file.write("/CapHeight %d\n" % glyphinfo_h[5])
            file.write("/StemV %d\n" % (glyphinfo_period[4]-glyphinfo_period[2]))

    def getglyphinfo(self, glyph, flex=True):
        logger.warning("We are about to extract font information for the Type 1 font '%s' from its pfb file. This is bad practice (and it's slow). You should use an afm file instead." % self.name)
        context = T1context(self, flex=flex)
        p = path()
        self.updateglyphpath(glyph, p, trafo.trafo(), context)
        bbox = p.bbox()
        return context.wx, context.wy, bbox.llx_pt, bbox.lly_pt, bbox.urx_pt, bbox.ury_pt

    def outputPFA(self, file, remove_UniqueID_lookup=False):
        """output the T1File in PFA format"""
        data1 = self.data1
        data3 = self.data3
        if remove_UniqueID_lookup:
            m1 = re.search(r"""FontDirectory\s*/%(name)s\s+known{/%(name)s\s+findfont\s+dup\s*/UniqueID\s+known\s*{\s*dup\s*
                              /UniqueID\s+get\s+\d+\s+eq\s+exch\s*/FontType\s+get\s+1\s+eq\s+and\s*}\s*{\s*pop\s+false\s*}\s*ifelse\s*
                              {save\s+true\s*}\s*{\s*false\s*}\s*ifelse\s*}\s*{\s*false\s*}\s*ifelse""" % {"name": self.name},
                           data1, re.VERBOSE)
            m3 = re.search(r"\s*{restore}\s*if", data3)
            if m1 and m3:
                data1 = data1[:m1.start()] + data1[m1.end():]
                data3 = data3[:m3.start()] + data3[m3.end():]
        file.write(data1)
        data2eexechex = binascii.b2a_hex(self.getdata2eexec())
        linelength = 64
        for i in range((len(data2eexechex)-1)//linelength + 1):
            file.write_bytes(data2eexechex[i*linelength: i*linelength+linelength])
            file.write("\n")
        file.write(data3)

    def outputPFB(self, file):
        """output the T1File in PFB format"""
        data2eexec = self.getdata2eexec()
        def pfblength(data):
            l = len(data)
            l, x1 = divmod(l, 256)
            l, x2 = divmod(l, 256)
            x4, x3 = divmod(l, 256)
            return chr(x1) + chr(x2) + chr(x3) + chr(x4)
        file.write("\200\1")
        file.write(pfblength(self.data1))
        file.write(self.data1)
        file.write("\200\2")
        file.write(pfblength(data2eexec))
        file.write(data2eexec)
        file.write("\200\1")
        file.write(pfblength(self.data3))
        file.write(self.data3)
        file.write("\200\3")

    def outputPS(self, file, writer):
        """output the PostScript code for the T1File to the file file"""
        self.outputPFA(file, remove_UniqueID_lookup=True)

    def outputPDF(self, file, writer):
        data2eexec = self.getdata2eexec()
        data3 = self.data3
        # we might be allowed to skip the third part ...
        if (data3.replace("\n", "")
                 .replace("\r", "")
                 .replace("\t", "")
                 .replace(" ", "")) == "0"*512 + "cleartomark":
            data3 = ""

        data = self.data1.encode("ascii", errors="surrogateescape") + data2eexec + data3.encode("ascii", errors="surrogateescape")
        if writer.compress and haszlib:
            data = zlib.compress(data)

        file.write("<<\n"
                   "/Length %d\n"
                   "/Length1 %d\n"
                   "/Length2 %d\n"
                   "/Length3 %d\n" % (len(data), len(self.data1), len(data2eexec), len(data3)))
        if writer.compress and haszlib:
            file.write("/Filter /FlateDecode\n")
        file.write(">>\n"
                   "stream\n")
        file.write_bytes(data)
        file.write("\n"
                   "endstream\n")

    @classmethod
    def from_PFA_bytes(cls, bytes):
        """create a T1File instance from a string of bytes corresponding to a PFA file"""
        try:
            m1 = bytes.index("eexec") + 6
            m2 = bytes.index("0"*40)
        except ValueError:
           raise FontFormatError

        data1 = bytes[:m1].decode("ascii", errors="surrogateescape")
        data2eexec = binascii.a2b_hex(bytes[m1: m2].replace(" ", "").replace("\r", "").replace("\n", ""))
        data3 = bytes[m2:].decode("ascii", errors="surrogateescape")
        return cls(data1, data2eexec, data3)

    @classmethod
    def from_PFA_filename(cls, filename):
        """create a T1File instance from PFA font file of given name"""
        with open(filename, "rb") as file:
            t1file = cls.from_PFA_bytes(file.read())
        return t1file

    @classmethod
    def from_PFB_bytes(cls, bytes):
        """create a T1File instance from a string of bytes corresponding to a PFB file"""

        def pfblength(s):
            if len(s) != 4:
                raise ValueError("invalid string length")
            return (s[0] +
                    s[1]*256 +
                    s[2]*256*256 +
                    s[3]*256*256*256)
        class consumer:
            def __init__(self, bytes):
                self.bytes = bytes
                self.pos = 0
            def __call__(self, n):
                result = self.bytes[self.pos:self.pos+n]
                self.pos += n
                return result

        consume = consumer(bytes)
        mark = consume(2)
        if mark != b"\200\1":
            raise FontFormatError
        data1 = consume(pfblength(consume(4))).decode("ascii", errors="surrogateescape")
        mark = consume(2)
        if mark != b"\200\2":
            raise FontFormatError
        data2eexec = b""
        while mark == b"\200\2":
            data2eexec = data2eexec + consume(pfblength(consume(4)))
            mark = consume(2)
        if mark != b"\200\1":
            raise FontFormatError
        data3 = consume(pfblength(consume(4))).decode("ascii", errors="surrogateescape")
        mark = consume(2)
        if mark != b"\200\3":
            raise FontFormatError
        if consume(1):
            raise FontFormatError

        return cls(data1, data2eexec, data3)

    @classmethod
    def from_PFB_filename(cls, filename):
        """create a T1File instance from PFB font file of given name"""
        with open(filename, "rb") as file:
            t1file = cls.from_PFB_bytes(file.read())
        return t1file

    @classmethod
    def from_PF_bytes(cls, bytes):
        try:
            return cls.from_PFB_bytes(bytes)
        except FontFormatError:
            return cls.from_PFA_bytes(bytes)

    @classmethod
    def from_PF_filename(cls, filename):
        """create a T1File instance from PFA or PFB font file of given name"""
        with open(filename, "rb") as file:
            t1file = cls.from_PF_bytes(file.read())
        return t1file



from pyx import *

text.set(text.LatexEngine, texenc='utf-8')
text.preamble(r'\usepackage[utf8]{inputenc}')

c = canvas.canvas()
c.text(0, 0, r'Héllò, wørłd!')
c.writeEPSfile()
c.writePDFfile()
c.writeSVGfile()



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2004, 2022 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2003-2004 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2004, 2022 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from fractions import Fraction

from pyx import text, utils
from pyx.graph.axis.tick import tick as Tick


class _texter:
    def labels(self, ticks):
        """fill the label attribute of ticks
        - ticks is a list of instances of tick
        - for each element of ticks the value of the attribute label is set to
          a string or MultiEngineText instance appropriate to the attributes
          num and denom of that tick instance
        - label attributes of the tick instances are just kept, whenever they
          are not equal to None
        - the method might modify the labelattrs attribute of the ticks; be sure
          to not modify it in-place!"""
        raise NotImplementedError


class decimal(_texter):
    r"a texter creating decimal labels (e.g. '1.234' or even '0.\overline{3}')"

    def __init__(self, prefix="", infix="", suffix="", equalprecision=False,
                       decimalsep=".", thousandsep="", thousandthpartsep="",
                       plus="", minus="-", period=r"\overline{%s}",
                       labelattrs=[text.mathmode]):
        r"""initializes the instance
        - prefix, infix, and suffix (strings) are added at the begin,
          immediately after the minus, and at the end of the label,
          respectively
        - equalprecision forces the same number of digits after decimalsep,
          even when the tailing digits are zero
        - decimalsep, thousandsep, and thousandthpartsep (strings)
          are used as separators
        - plus or minus (string) is inserted for non-negative or negative numbers
        - period (string) is taken as a format string generating a period;
          it has to contain exactly one string insert operators "%s" for the
          period; usually it should be r"\overline{%s}"
        - labelattrs is a list of attributes to be added to the label attributes
          given in the painter"""
        self.prefix = prefix
        self.infix = infix
        self.suffix = suffix
        self.equalprecision = equalprecision
        self.decimalsep = decimalsep
        self.thousandsep = thousandsep
        self.thousandthpartsep = thousandthpartsep
        self.plus = plus
        self.minus = minus
        self.period = period
        self.labelattrs = labelattrs

    def __call__(self, **kwargs):
        return decimal(**utils.merge_members_kwargs(self, [kwargs],
                                                    ["prefix", "infix", "suffix", "equalprecision",
                                                     "decimalsep", "thousandsep", "thousandthpartsep", "plus",
                                                     "minus", "period", "labelattrs"]))

    def labels(self, ticks):
        labeledticks = []
        maxdecprecision = 0
        for tick in ticks:
            if tick.label is None and tick.labellevel is not None:
                labeledticks.append(tick)
                m, n = tick.num, tick.denom
                if m < 0: m = -m
                if n < 0: n = -n
                quotient, remainder = divmod(m, n)
                quotient = str(quotient)
                if len(self.thousandsep):
                    l = len(quotient)
                    tick.label = ""
                    for i in range(l):
                        tick.label += quotient[i]
                        if not ((l-i-1) % 3) and l > i+1:
                            tick.label += self.thousandsep
                else:
                    tick.label = quotient
                if remainder:
                    tick.label += self.decimalsep
                oldremainders = []
                tick.temp_decprecision = 0
                while (remainder):
                    tick.temp_decprecision += 1
                    if remainder in oldremainders:
                        tick.temp_decprecision = None
                        periodstart = len(tick.label) - (len(oldremainders) - oldremainders.index(remainder))
                        tick.label = tick.label[:periodstart] + self.period % tick.label[periodstart:]
                        break
                    oldremainders += [remainder]
                    remainder *= 10
                    quotient, remainder = divmod(remainder, n)
                    if not ((tick.temp_decprecision - 1) % 3) and tick.temp_decprecision > 1:
                        tick.label += self.thousandthpartsep
                    tick.label += str(quotient)
                else:
                    if maxdecprecision < tick.temp_decprecision:
                        maxdecprecision = tick.temp_decprecision
        if self.equalprecision:
            for tick in labeledticks:
                if tick.temp_decprecision is not None:
                    if tick.temp_decprecision == 0 and maxdecprecision > 0:
                        tick.label += self.decimalsep
                    for i in range(tick.temp_decprecision, maxdecprecision):
                        if not ((i - 1) % 3) and i > 1:
                            tick.label += self.thousandthpartsep
                        tick.label += "0"
        for tick in labeledticks:
            if tick.num * tick.denom < 0:
                plusminus = self.minus
            else:
                plusminus = self.plus
            tick.label = "%s%s%s%s%s" % (self.prefix, plusminus, self.infix, tick.label, self.suffix)
            tick.labelattrs = tick.labelattrs + self.labelattrs

            # del tick.temp_decprecision  # we've inserted this temporary variable ... and do not care any longer about it



class skipmantissaunity:
    pass

skipmantissaunity.never = 0
skipmantissaunity.each = 1
skipmantissaunity.all = 2


class default(_texter):

    r"a texter creating regular (e.g. '2') and exponential (e.g. '2\cdot10^5') labels"

    def __init__(self, multiplicationtex=r"\cdot{}", multiplicationunicode="·", base=Fraction(10),
                       skipmantissaunity=skipmantissaunity.all, minusunity="-",
                       minexponent=4, minnegexponent=None, uniformexponent=True,
                       mantissatexter=decimal(), basetexter=decimal(), exponenttexter=decimal(),
                       labelattrs=[text.mathmode], 
                       **kwargs):
        r"""initializes the instance
        - multiplicationtex and multiplicationunicode are the strings to
          indicate the multiplication between the mantissa and the base
          number for the TexEngine and the UnicodeEngine, respecitvely
        - base is the number of the base of the exponent
        - skipmantissaunity is either skipmantissaunity.never (never skip the
          unity mantissa), skipmantissaunity.each (skip the unity mantissa
          whenever it occurs for each label separately), or skipmantissaunity.all
          (skip the unity mantissa whenever if all labels happen to be
          mantissafixed with unity)
        - minusunity is used as the output of -unity for the mantissa
        - minexponent is the minimal positive exponent value to be printed by
          exponential notation
        - minnegexponent is the minimal negative exponent value to be printed by
          exponential notation, for None it is considered to be equal to minexponent
        - uniformexponent forces all numbers to be written in exponential notation
          when at least one label excets the limits for non-exponential
          notiation
        - mantissatexter, basetexter, and exponenttexter generate the texts
          for the mantissa, basetexter, and exponenttexter
        - labelattrs is a list of attributes to be added to the label attributes
          given in the painter"""
        split_kwargs, rest_kwargs = utils.kwsplit(kwargs, ["mantissatexter", "basetexter", "exponenttexter"], allow_only_split=True)
        if "mantissatexter" in split_kwargs: mantissatexter = decimal(**split_kwargs["mantissatexter"])
        if "basetexter" in split_kwargs: basetexter = decimal(**split_kwargs["basetexter"])
        if "exponenttexter" in split_kwargs: texter = decimal(**split_kwargs["exponenttexter"])

        self.multiplicationtex = multiplicationtex
        self.multiplicationunicode = multiplicationunicode
        self.base = base
        self.skipmantissaunity = skipmantissaunity
        self.minusunity = minusunity
        self.minexponent = minexponent
        self.minnegexponent = minnegexponent if minnegexponent is not None else minexponent
        self.uniformexponent = uniformexponent
        self.labelattrs = labelattrs
        self.mantissatexter = mantissatexter
        self.basetexter = basetexter
        self.exponenttexter = exponenttexter
        self.kwargs = kwargs

    def __call__(self, **kwargs):
        return default(**utils.merge_members_kwargs(self, [self.kwargs, kwargs],
                                                    ["multiplicationtex", "multiplicationunicode", "base", "skipmantissaunity",
                                                     "minusunity", "minexponent", "minnegexponent", "uniformexponent",
                                                     "labelattrs", "mantissatexter", "basetexter", "exponenttexter"]))

    def labels(self, ticks):
        labeledticks = []
        for tick in ticks:
            if tick.label is None and tick.labellevel is not None:
                labeledticks.append(tick)

                tick.labelattrs = tick.labelattrs + self.labelattrs

                if tick.num:
                    # express tick = tick.temp_sign * tick.temp_mantissa * self.base ** tick.temp_exponent with 1 <= temp_mantissa < self.base 
                    # and decide whether a tick is to be written in exponential notation
                    tick.temp_sign = 1 if tick >= 0 else -1
                    tick.temp_mantissa = abs(Fraction(tick.num, tick.denom))
                    tick.temp_exponent = 0
                    while tick.temp_mantissa >= self.base:
                        tick.temp_exponent += 1
                        tick.temp_mantissa /= self.base
                    while tick.temp_mantissa < 1:
                        tick.temp_exponent -= 1
                        tick.temp_mantissa *= self.base
                    tick.temp_wantexponent = not (-self.minnegexponent < tick.temp_exponent < self.minexponent)
                else:
                    tick.temp_mantissa = tick.temp_exponent = 0
                    tick.temp_sign = 1
                    tick.temp_wantexponent = not (-self.minnegexponent < 0 < self.minexponent)

        # make decision on exponential notation uniform if requested
        if self.uniformexponent and any(tick.temp_wantexponent for tick in labeledticks):
            for tick in labeledticks:
                if tick.num:
                    tick.temp_wantexponent = True

        # mark mantissa == 1 to be not labeled
        if self.skipmantissaunity == skipmantissaunity.each:
            for tick in labeledticks:
                if tick.temp_wantexponent and tick.temp_mantissa == 1:
                    tick.temp_mantissa = None
        elif self.skipmantissaunity == skipmantissaunity.all and all(tick.temp_mantissa == 1 for tick in labeledticks if tick.temp_wantexponent):
            for tick in labeledticks:
                if tick.temp_wantexponent:
                    tick.temp_mantissa = None

        # construct labels
        basetick = Tick(self.base, labellevel=0)
        self.basetexter.labels([basetick])
        for tick in labeledticks:
            if tick.temp_wantexponent:
                if tick.temp_mantissa is not None:
                    tick.temp_mantissatick = Tick(tick.temp_sign * tick.temp_mantissa, labellevel=0)
                tick.temp_exponenttick = Tick(tick.temp_exponent, labellevel=0)
            else:
                tick.temp_mantissatick = tick

        self.mantissatexter.labels([tick.temp_mantissatick for tick in labeledticks if tick.temp_mantissa is not None])
        self.exponenttexter.labels([tick.temp_exponenttick for tick in labeledticks if tick.temp_wantexponent])
        for tick in labeledticks:
            if tick.temp_wantexponent:
                if tick.temp_mantissa is not None:
                    mantissalabel_tex = tick.temp_mantissatick.label + self.multiplicationtex
                    mantissalabel_unicode = tick.temp_mantissatick.label + self.multiplicationunicode
                else:
                    mantissalabel_tex = self.minusunity if tick.temp_sign == -1 else ""
                    mantissalabel_unicode = self.minusunity if tick.temp_sign == -1 else ""
                tick.label = text.MultiEngineText("%s%s^{%s}" % (mantissalabel_tex, basetick.label, tick.temp_exponenttick.label), [mantissalabel_unicode + basetick.label, text.Text(tick.temp_exponenttick.label, scale=0.8, shift=0.5)])


class rational(_texter):
    r"a texter creating rational labels (e.g. 'a/b' or even 'a \over b')"
    # we use divmod here to be more explicit

    def __init__(self, prefix="", infix="", suffix="",
                       numprefix="", numinfix="", numsuffix="",
                       denomprefix="", denominfix="", denomsuffix="",
                       plus="", minus="-", minuspos=0, over=r"{{%s}\over{%s}}",
                       equaldenom=False, skip1=True, skipnum0=True, skipnum1=True, skipdenom1=True,
                       labelattrs=[text.mathmode]):
        r"""initializes the instance
        - prefix, infix, and suffix (strings) are added at the begin,
          immediately after the minus, and at the end of the label,
          respectively
        - prefixnum, infixnum, and suffixnum (strings) are added
          to the labels numerator correspondingly
        - prefixdenom, infixdenom, and suffixdenom (strings) are added
          to the labels denominator correspondingly
        - plus or minus (string) is inserted for non-negative or negative numbers
        - minuspos is an integer, which determines the position, where the
          plus or minus sign has to be placed; the following values are allowed:
            1 - writes the plus or minus in front of the numerator
            0 - writes the plus or minus in front of the hole fraction
           -1 - writes the plus or minus in front of the denominator
        - over (string) is taken as a format string generating the
          fraction bar; it has to contain exactly two string insert
          operators "%s" -- the first for the numerator and the second
          for the denominator; by far the most common examples are
          r"{{%s}\over{%s}}" and "{{%s}/{%s}}"
        - usually the numerator and denominator are canceled; however,
          when equaldenom is set, the least common multiple of all
          denominators is used
        - skip1 (boolean) just prints the prefix, the plus or minus,
          the infix and the suffix, when the value is plus or minus one
          and at least one of prefix, infix and the suffix is present
        - skipnum0 (boolean) just prints a zero instead of
          the hole fraction, when the numerator is zero;
          no prefixes, infixes, and suffixes are taken into account
        - skipnum1 (boolean) just prints the numprefix, the plus or minus,
          the numinfix and the numsuffix, when the num value is plus or minus one
          and at least one of numprefix, numinfix and the numsuffix is present
        - skipdenom1 (boolean) just prints the numerator instead of
          the hole fraction, when the denominator is one and none of the parameters
          denomprefix, denominfix and denomsuffix are set and minuspos is not -1 or the
          fraction is positive
        - labelattrs is a list of attributes for a textengines text method;
          None is considered as an empty list; labelattrs might be changed
          in the painter as well"""
        # XXX should get separate texters for nominator and denominator
        self.prefix = prefix
        self.infix = infix
        self.suffix = suffix
        self.numprefix = numprefix
        self.numinfix = numinfix
        self.numsuffix = numsuffix
        self.denomprefix = denomprefix
        self.denominfix = denominfix
        self.denomsuffix = denomsuffix
        self.plus = plus
        self.minus = minus
        self.minuspos = minuspos
        self.over = over
        self.equaldenom = equaldenom
        self.skip1 = skip1
        self.skipnum0 = skipnum0
        self.skipnum1 = skipnum1
        self.skipdenom1 = skipdenom1
        self.labelattrs = labelattrs

    def __call__(self, **kwargs):
        return rational(**utils.merge_members_kwargs(self, [kwargs],
                                                    ["prefix", "infix", "suffix", "numprefix", "numinfix", "numsuffix",
                                                     "denomprefix", "denominfix", "denomsuffix", "plus", "minus", "minuspos",
                                                     "over", "equaldenom", "skip1", "skipnum0", "skinum1", "skipdenom1", "labelattrs"]))

    def gcd(self, *n):
        """returns the greates common divisor of all elements in n
        - the elements of n must be non-negative integers
        - return None if the number of elements is zero
        - the greates common divisor is not affected when some
          of the elements are zero, but it becomes zero when
          all elements are zero"""
        if len(n) == 2:
            i, j = n
            if i < j:
                i, j = j, i
            while j > 0:
                i, (dummy, j) = j, divmod(i, j)
            return i
        if len(n):
            res = n[0]
            for i in n[1:]:
                res = self.gcd(res, i)
            return res

    def lcm(self, *n):
        """returns the least common multiple of all elements in n
        - the elements of n must be non-negative integers
        - return None if the number of elements is zero
        - the least common multiple is zero when some of the
          elements are zero"""
        if len(n):
            res = n[0]
            for i in n[1:]:
                res = divmod(res * i, self.gcd(res, i))[0]
            return res

    def labels(self, ticks):
        labeledticks = []
        for tick in ticks:
            if tick.label is None and tick.labellevel is not None:
                labeledticks.append(tick)
                tick.temp_rationalnum = tick.num
                tick.temp_rationaldenom = tick.denom
                tick.temp_rationalminus = 1
                if tick.temp_rationalnum < 0:
                    tick.temp_rationalminus = -tick.temp_rationalminus
                    tick.temp_rationalnum = -tick.temp_rationalnum
                if tick.temp_rationaldenom < 0:
                    tick.temp_rationalminus = -tick.temp_rationalminus
                    tick.temp_rationaldenom = -tick.temp_rationaldenom
                gcd = self.gcd(tick.temp_rationalnum, tick.temp_rationaldenom)
                (tick.temp_rationalnum, dummy1), (tick.temp_rationaldenom, dummy2) = divmod(tick.temp_rationalnum, gcd), divmod(tick.temp_rationaldenom, gcd)
        if self.equaldenom:
            equaldenom = self.lcm(*[tick.temp_rationaldenom for tick in ticks if tick.label is None])
            if equaldenom is not None:
                for tick in labeledticks:
                    factor, dummy = divmod(equaldenom, tick.temp_rationaldenom)
                    tick.temp_rationalnum, tick.temp_rationaldenom = factor * tick.temp_rationalnum, factor * tick.temp_rationaldenom
        for tick in labeledticks:
            rationalminus = rationalnumminus = rationaldenomminus = ""
            if tick.temp_rationalminus == -1:
                plusminus = self.minus
            else:
                plusminus = self.plus
            if self.minuspos == 0:
                rationalminus = plusminus
            elif self.minuspos == 1:
                rationalnumminus = plusminus
            elif self.minuspos == -1:
                rationaldenomminus = plusminus
            else:
                raise RuntimeError("invalid minuspos")
            if self.skipnum0 and tick.temp_rationalnum == 0:
                tick.label = "0"
            elif (self.skip1 and self.skipdenom1 and tick.temp_rationalnum == 1 and tick.temp_rationaldenom == 1 and
                  (len(self.prefix) or len(self.infix) or len(self.suffix)) and
                  not len(rationalnumminus) and not len(self.numprefix) and not len(self.numinfix) and not len(self.numsuffix) and
                  not len(rationaldenomminus) and not len(self.denomprefix) and not len(self.denominfix) and not len(self.denomsuffix)):
                tick.label = "%s%s%s%s" % (self.prefix, rationalminus, self.infix, self.suffix)
            else:
                if self.skipnum1 and tick.temp_rationalnum == 1 and (len(self.numprefix) or len(self.numinfix) or len(self.numsuffix)):
                    tick.temp_rationalnum = "%s%s%s%s" % (self.numprefix, rationalnumminus, self.numinfix, self.numsuffix)
                else:
                    tick.temp_rationalnum = "%s%s%s%i%s" % (self.numprefix, rationalnumminus, self.numinfix, tick.temp_rationalnum, self.numsuffix)
                if self.skipdenom1 and tick.temp_rationaldenom == 1 and not len(rationaldenomminus) and not len(self.denomprefix) and not len(self.denominfix) and not len(self.denomsuffix):
                    tick.label = "%s%s%s%s%s" % (self.prefix, rationalminus, self.infix, tick.temp_rationalnum, self.suffix)
                else:
                    tick.temp_rationaldenom = "%s%s%s%i%s" % (self.denomprefix, rationaldenomminus, self.denominfix, tick.temp_rationaldenom, self.denomsuffix)
                    tick.label = text.MultiEngineText("%s%s%s%s%s" % (self.prefix, rationalminus, self.infix, self.over % (tick.temp_rationalnum, tick.temp_rationaldenom), self.suffix),
                                                      ["%s%s%s" % (self.prefix, rationalminus, self.infix)] + [text.StackedText([text.Text(tick.temp_rationalnum, shift=0.3), text.Text(tick.temp_rationaldenom, shift=-0.9)], frac=True, align=0.5)] + [self.suffix])
            tick.labelattrs = tick.labelattrs + self.labelattrs

            # del tick.temp_rationalnum    # we've inserted those temporary variables ... and do not care any longer about them
            # del tick.temp_rationaldenom
            # del tick.temp_rationalminus




from pyx import *

c = canvas.canvas()
circ = path.circle(0, 0, 1)

# variant 1: use trafo as a deformer
c.stroke(circ, [style.linewidth.THIck, 
                trafo.scale(sx=2, sy=0.9), trafo.rotate(45), trafo.translate(1, 0)])

# variant 2: transform a subcanvas
sc = canvas.canvas()
sc.stroke(circ, [style.linewidth.THIck])
c.insert(sc, [trafo.scale(sx=2, sy=0.9), trafo.rotate(45), trafo.translate(5, 0)])

c.writeEPSfile("ellipse")
c.writePDFfile("ellipse")
c.writeSVGfile("ellipse")
